"""
This file contains example Mixins.

Mixins can add reusable fields and behavior (optimally both, otherwise it doesn't add much).
"""

import base64
import datetime as dt
import os
import typing as t
import warnings

from pydal import DAL
from pydal.validators import IS_NOT_IN_DB, ValidationError
from slugify import slugify

from .core import TypeDAL
from .fields import DatetimeField, StringField
from .tables import _TypedTable
from .types import OpRow, Set, T_MetaInstance

if t.TYPE_CHECKING:
    from .tables import TypedTable  # noqa: F401


class Mixin(_TypedTable):
    """
    A mixin should be derived from this class.

    The mixin base class itself doesn't do anything,
    but using it makes sure the mixin fields are placed AFTER the table's normal fields (instead of before)

    During runtime, mixin should not have a base class in order to prevent MRO issues
        ('inconsistent method resolution' or 'metaclass conflicts')
    """

    __settings__: t.ClassVar[dict[str, t.Any]]

    def __init_subclass__(cls, **kwargs: t.Any):
        """
        Ensures __settings__ exists for other mixins.
        """
        cls.__settings__ = getattr(cls, "__settings__", None) or {}


class TimestampsMixin(Mixin):
    """
    A Mixin class for adding timestamp fields to a model.
    """

    created_at = DatetimeField(default=dt.datetime.now, writable=False)
    updated_at = DatetimeField(default=dt.datetime.now, writable=False)

    @classmethod
    def __on_define__(cls, db: TypeDAL) -> None:
        """
        Hook called when defining the model to initialize timestamps.

        Args:
            db (TypeDAL): The database layer.
        """
        super().__on_define__(db)

        def set_updated_at(_: Set, row: OpRow) -> None:
            """
            Callback function to update the 'updated_at' field before saving changes.

            Args:
                _: Set: Unused parameter.
                row (OpRow): The row to update.
            """
            row["updated_at"] = dt.datetime.now()

        cls._before_update.append(set_updated_at)


def slug_random_suffix(length: int = 8) -> str:
    """
    Generate a random suffix to make slugs unique, even when titles are the same.

    UUID4 uses 16 bytes, but 8 is probably more than enough given you probably don't have THAT much duplicate titles.
    Strip away '=' to make it URL-safe
        (even though 'urlsafe_b64encode' sounds like it should already be url-safe - it is not)
    """
    return base64.urlsafe_b64encode(os.urandom(length)).rstrip(b"=").decode().strip("=")


T = t.TypeVar("T")


# noinspection PyPep8Naming
class HAS_UNIQUE_SLUG(IS_NOT_IN_DB):
    """
    Checks if slugified field is already in the db.

    Usage:
        table.name = HAS_UNIQUE_SLUG(db, "table.slug")
    """

    def __init__(
        self,
        db: TypeDAL | DAL,
        field: str,  # table.slug
        error_message: str = "This slug is not unique: %s.",
    ):
        """
        Based on IS_NOT_IN_DB but with less options and a different default error message.
        """
        super().__init__(db, field, error_message)

    def validate(self, original: T, record_id: t.Optional[int] = None) -> T:
        """
        Performs checks to see if the slug already exists for a different row.
        """
        value = slugify(str(original))
        if not value.strip():
            raise ValidationError(self.translator(self.error_message))

        (tablename, fieldname) = str(self.field).split(".")
        table = self.dbset.db[tablename]
        field = table[fieldname]
        query = field == value

        # make sure exclude the record_id
        row_id = record_id or self.record_id
        if isinstance(row_id, dict):  # pragma: no cover
            row_id = table(**row_id)
        if row_id is not None:
            query &= table._id != row_id
        subset = self.dbset(query)

        if subset.count():
            raise ValidationError(self.error_message % value)

        return original


class SlugMixin(Mixin):
    """
    (Opinionated) example mixin to add a 'slug' field, which depends on a user-provided other field.

    Some random bytes are added at the end to prevent duplicates.

    Example:
        >>> class MyTable(TypedTable, SlugMixin, slug_field="some_name", slug_suffix_length=8):
        >>>    some_name: str
        >>>    ...
    """

    # pub:
    slug = StringField(unique=True, writable=False)
    # priv:
    __settings__: t.TypedDict(  # type: ignore
        "SlugFieldSettings",
        {
            "slug_field": str,
            "slug_suffix": int,
        },
    )  # set via init subclass

    def __init_subclass__(
        cls,
        slug_field: t.Optional[str] = None,
        slug_suffix_length: int = 0,
        slug_suffix: t.Optional[int] = None,
        **kw: t.Any,
    ) -> None:
        """
        Bind 'slug field' option to be used later (on_define).

        You can control the length of the random suffix with the `slug_suffix_length` option (0 is no suffix).
        """
        super().__init_subclass__(**kw)

        # unfortunately, PyCharm and mypy do not recognize/autocomplete/typecheck init subclass (keyword) arguments.
        if slug_field is None:
            raise ValueError(
                "SlugMixin requires a valid slug_field setting: "
                "e.g. `class MyClass(TypedTable, SlugMixin, slug_field='title'): ...`",
            )

        if slug_suffix:
            warnings.warn(
                "The 'slug_suffix' option is deprecated, use 'slug_suffix_length' instead.",
                DeprecationWarning,
            )

        slug_suffix = slug_suffix_length or kw.get("slug_suffix", 0)

        # append settings:
        cls.__settings__["slug_field"] = slug_field
        cls.__settings__["slug_suffix"] = slug_suffix

    @classmethod
    def __generate_slug_before_insert(cls, row: OpRow) -> None:
        if row.get("slug"):  # type: ignore
            # manually set -> skip
            return None

        settings = cls.__settings__

        text_input = row[settings["slug_field"]]
        generated_slug = slugify(text_input)

        if suffix_len := settings["slug_suffix"]:
            generated_slug += f"-{slug_random_suffix(suffix_len)}"

        row["slug"] = slugify(generated_slug)
        return None

    @classmethod
    def __on_define__(cls, db: TypeDAL) -> None:
        """
        When db is available, include a before_insert hook to generate and include a slug.
        """
        super().__on_define__(db)
        settings = cls.__settings__

        # slugs should not be editable (for SEO reasons), so there is only a before insert hook:
        cls._before_insert.append(cls.__generate_slug_before_insert)

        if settings["slug_suffix"] == 0:
            # add a validator to the field that will be slugified:
            slug_field = getattr(cls, settings["slug_field"])
            current_requires = getattr(slug_field, "requires", None) or []
            if not isinstance(current_requires, list):
                current_requires = [current_requires]

            current_requires.append(HAS_UNIQUE_SLUG(db, f"{cls}.slug"))

            slug_field.requires = current_requires

    @classmethod
    def from_slug(cls: t.Type[T_MetaInstance], slug: str, join: bool = True) -> t.Optional[T_MetaInstance]:
        """
        Find a row by its slug.
        """
        builder = cls.where(slug=slug)
        if join:
            builder = builder.join()

        return builder.first()

    @classmethod
    def from_slug_or_fail(cls: t.Type[T_MetaInstance], slug: str, join: bool = True) -> T_MetaInstance:
        """
        Find a row by its slug, or raise an error if it doesn't exist.
        """
        builder = cls.where(slug=slug)
        if join:
            builder = builder.join()

        return builder.first_or_fail()



"""
Contains base functionality related to Relationships.
"""

import inspect
import typing as t
import warnings

import pydal.objects

from .config import LazyPolicy
from .constants import JOIN_OPTIONS
from .core import TypeDAL
from .fields import TypedField
from .helpers import extract_type_optional, looks_like, unwrap_type
from .types import Condition, OnQuery, T_Field

To_Type = t.TypeVar("To_Type")


# default lazy policy is defined at the TypeDAL() instance settings level


class Relationship(t.Generic[To_Type]):
    """
    Define a relationship to another table.
    """

    _type: t.Type[To_Type]
    table: t.Type["TypedTable"] | type | str  # use get_table() to resolve later on
    condition: Condition
    condition_and: Condition
    on: OnQuery
    multiple: bool
    join: JOIN_OPTIONS
    _lazy: LazyPolicy | None
    nested: dict[str, t.Self]
    explicit: bool
    name: str | None = None  # set by __set_name__

    def __init__(
        self,
        _type: t.Type[To_Type],
        condition: Condition = None,
        join: JOIN_OPTIONS = None,
        on: OnQuery = None,
        condition_and: Condition = None,
        nested: dict[str, t.Self] = None,
        lazy: LazyPolicy | None = None,
        explicit: bool = False,
    ):
        """
        Should not be called directly, use relationship() instead!
        """
        if condition and on:
            raise self._error_duplicate_condition(condition, on)

        self._type = _type
        self.condition = condition
        self.join = "left" if on else join  # .on is always left join!
        self.on = on
        self.condition_and = condition_and
        self._lazy = lazy

        if args := t.get_args(_type):
            self.table = unwrap_type(args[0])
            self.multiple = True
        else:
            self.table = t.cast(type[TypedTable], _type)
            self.multiple = False

        if isinstance(self.table, str):
            self.table = TypeDAL.to_snake(self.table)

        self.explicit = explicit
        self.nested = nested or {}

    def clone(self, **update: t.Any) -> "Relationship[To_Type]":
        """
        Create a copy of the relationship, possibly updated.
        """
        condition = update.get("condition")
        on = update.get("on")

        if on and condition:  # pragma: no cover
            raise self._error_duplicate_condition(condition, on)

        return self.__class__(
            update.get("_type") or self._type,
            None if on else (condition or self.condition),
            update.get("join") or self.join,
            None if condition else (on or self.on),
            update.get("condition_and") or self.condition_and,
            (self.nested | extra) if (extra := update.get("nested")) else self.nested,  # type: ignore
            update.get("lazy") or self._lazy,
        )

    @staticmethod
    def _error_duplicate_condition(condition: Condition, on: OnQuery) -> t.Never:
        warnings.warn(f"Relation | Both specified! {condition=} {on=}")
        raise ValueError("Please specify either a condition or an 'on' statement for this relationship!")

    def __repr__(self) -> str:
        """
        Representation of the relationship.
        """
        if callback := self.condition or self.on:
            src_code = inspect.getsource(callback).strip()

            if c_and := self.condition_and:
                and_code = inspect.getsource(c_and).strip()
                src_code += " AND " + and_code
        else:
            cls_name = self._type if isinstance(self._type, str) else self._type.__name__
            src_code = f"to {cls_name} (missing condition)"

        join = f":{self.join}" if self.join else ""
        lazy_str = f" lazy={self.lazy}" if self.lazy != "warn" else ""
        return f"<Relationship{join}{lazy_str} {src_code}>"

    def __set_name__(self, owner: t.Type["TypedTable"], name: str) -> None:
        """Called automatically when assigned to a class attribute."""
        self.name = name

    def get_table(self, db: "TypeDAL") -> t.Type["TypedTable"]:
        """
        Get the table this relationship is bound to.
        """
        table = self.table  # can be a string because db wasn't available yet

        if isinstance(table, str):
            if mapped := db._class_map.get(table):
                # yay
                return mapped

            # boo, fall back to untyped table but pretend it is typed:
            return t.cast(t.Type["TypedTable"], db[table])  # eh close enough!

        return table

    def get_db(self) -> TypeDAL | None:
        """
        Retrieves the database instance associated with the table.

        Returns:
            TypeDAL | None: The database instance if it exists, or None otherwise.
        """
        return getattr(self.table, "_db", None)

    @property
    def lazy(self) -> LazyPolicy:
        """
        Gets the lazy policy configured in the current context.

        The method first checks for a customized lazy policy for this relationship.
        If not found, it attempts to retrieve the lazy policy from the database.
        If neither option is available, it returns a conservative fallback value.

        Returns:
            LazyPolicy or str: The configured lazy policy or a fallback value.
        """
        if customized := self._lazy:
            return customized

        if db := self.get_db():
            return db._config.lazy_policy

        # conservative fallback:
        return "warn"

    def get_table_name(self) -> str:
        """
        Get the name of the table this relationship is bound to.
        """
        if isinstance(self.table, str):
            return self.table

        if isinstance(self.table, pydal.objects.Table):
            return str(self.table)

        # else: typed table
        try:
            table = self.table._ensure_table_defined() if issubclass(self.table, TypedTable) else self.table
        except Exception:  # pragma: no cover
            table = self.table

        return str(table)

    def __get__(
        self,
        instance: "TypedTable",
        owner: t.Type["TypedTable"],
    ) -> "t.Optional[list[t.Any]] | Relationship[To_Type]":
        """
        Relationship is a descriptor class, which can be returned from a class but not an instance.

        For an instance, using .join() will replace the Relationship with the actual data.
        Behavior when accessed without joining depends on the lazy policy.
        """
        if not instance:
            # relationship queried on class, that's allowed
            return self

        # instance: TypedTable instance
        # owner: TypedTable class

        if not self.name:  # pragma: no cover
            raise ValueError("Relationship does not seem to be connected to a table field.")

        # Handle different lazy policies
        if self.lazy == "forbid":  # pragma: no cover
            raise AttributeError(
                f"Accessing relationship '{self.name}' without joining is forbidden. "
                f"Use .join('{self.name}') in your query or set lazy='allow' if this is intentional.",
            )

        fallback_value: t.Optional[list[t.Any]] = [] if self.multiple else None

        if self.lazy == "ignore":  # pragma: no cover
            # Return empty silently
            return fallback_value

        if self.lazy == "warn":
            # Warn and return empty
            warnings.warn(
                f"Trying to access relationship '{self.name}' without joining. "
                f"Did you forget to use .join('{self.name}')? Returning empty value.",
                category=RuntimeWarning,
            )
            return fallback_value

        # For "tolerate" and "allow", we fetch the data
        try:
            resolved_table = self.get_table(instance._db)

            builder = owner.where(id=instance.id).join(self.name)
            if issubclass(resolved_table, TypedTable) or isinstance(resolved_table, pydal.objects.Table):
                # is a table so we can select ALL and ignore non-required fields of parent row:
                builder = builder.select(owner.id, resolved_table.ALL)

            if self.lazy == "tolerate":
                warnings.warn(
                    f"Lazy loading relationship '{self.name}'. "
                    "This performs an extra database query. "
                    f"Consider using .join('{self.name}') for better performance.",
                    category=RuntimeWarning,
                )

            return builder.first()[self.name]  # type: ignore
        except Exception as e:  # pragma: no cover
            warnings.warn(
                f"Failed to lazy load relationship '{self.name}': {e}",
                category=RuntimeWarning,
                source=e,
            )

            return fallback_value


@t.overload
def relationship(
    _type: type[list[To_Type]],
    condition: Condition = None,
    join: JOIN_OPTIONS = None,
    on: OnQuery = None,
    lazy: LazyPolicy | None = None,
    explicit: bool = False,
) -> list[To_Type]:
    """
    Define a relationship that returns a list of related instances.

    Args:
        _type: A list type hint like list[Office] to indicate multiple related records.

    Returns:
        A list of related instances.
    """


@t.overload
def relationship(
    _type: t.Type[To_Type] | str,
    condition: Condition = None,
    *,
    join: t.Literal["inner"],
    on: OnQuery = None,
    lazy: LazyPolicy | None = None,
    explicit: bool = False,
) -> To_Type:
    """
    Define a relationship that returns a single related instance (never None with inner join).

    Args:
        _type: A type or string reference like City to indicate a single related record.
        join: Set to 'inner' to guarantee a non-null result.

    Returns:
        A single related instance (guaranteed non-null with inner join).
    """


@t.overload
def relationship(
    _type: t.Type[To_Type] | str,
    condition: Condition = None,
    join: JOIN_OPTIONS = None,
    on: OnQuery = None,
    lazy: LazyPolicy | None = None,
    explicit: bool = False,
) -> To_Type | None:
    """
    Define a relationship that returns a single optional related instance.

    Args:
        _type: A type or string reference like City to indicate a single related record.

    Returns:
        A single related instance or None.
    """


def relationship(
    _type: type[list[To_Type]] | t.Type[To_Type] | str,
    condition: Condition = None,
    join: JOIN_OPTIONS = None,
    on: OnQuery = None,
    lazy: LazyPolicy | None = None,
    explicit: bool = False,
) -> list[To_Type] | To_Type | None:
    """
    Define a relationship to another table, when its id is not stored in the current table.

    Args:
        _type: The type of the related table. Use list[Type] for one-to-many relationships.
        condition: Lambda function defining the join condition between tables.
                   Example: lambda self, post: self.id == post.author
        join: Join strategy ('left', 'inner', etc.). Defaults to 'left' when using 'on'.
              When 'inner' is used with a single type, the result is guaranteed non-null.
        on: Alternative to condition for complex queries with pivot tables.
            Allows specifying multiple join conditions to avoid cross joins.
        lazy: Controls behavior when accessing relationship data without explicitly joining:
            - "forbid": Raise an error (strictest, prevents N+1 queries)
            - "warn": Return empty value with warning
            - "ignore": Return empty value silently
            - "tolerate": Fetch data with warning (convenient but warns about performance)
            - "allow": Fetch data silently (most permissive, use only for known cheap queries)
        explicit: If True, this relationship is only joined when explicitly requested
                  (e.g. User.join("tags")). Bare User.join() calls will skip it.
                  Useful for expensive or rarely-needed relationships. Defaults to False.

    Example:
        class User(TypedTable):
            name: str

            posts = relationship(list["Post"], condition=lambda self, post: self.id == post.author, join='left')

        class Post(TypedTable):
            title: str
            author: User

    User.join("posts").first() # User instance with list[Post] in .posts

    Here, Post stores the User ID, but `relationship(list["Post"])` still allows you to get the user's posts.
    In this case, the join strategy is set to LEFT so users without posts are also still selected.

    For complex queries with a pivot table, 'on' can be set instead of 'condition':
        class User(TypedTable):
        ...

        tags = relationship(list["Tag"], on=lambda self, tag: [
                Tagged.on(Tagged.entity == entity.gid),
                Tag.on((Tagged.tag == tag.id)),
            ])

    If you'd try to capture this in a single 'condition', pydal would create a cross join which is much less efficient.
    """
    return t.cast(
        # note: The descriptor `Relationship[To_Type]` is more correct, but pycharm doesn't really get that.
        # so for ease of use, just cast to the refered type for now!
        # e.g. x = relationship(Author) -> x: Author
        To_Type,
        Relationship(_type, condition, join, on, lazy=lazy, explicit=explicit),  # type: ignore
    )


def _generate_relationship_condition(_: t.Type["TypedTable"], key: str, field: T_Field) -> Condition:
    origin = t.get_origin(field)
    # else: generic

    if origin is list:
        # field = typing.get_args(field)[0]  # actual field
        # return lambda _self, _other: cls[key].contains(field)

        return lambda _self, _other: _self[key].contains(_other.id)
    else:
        # normal reference
        # return lambda _self, _other: cls[key] == field.id
        return lambda _self, _other: _self[key] == _other.id


def to_relationship(
    cls: t.Type["TypedTable"] | type[t.Any],
    key: str,
    field: T_Field,
) -> t.Optional[Relationship[t.Any]]:
    """
    Used to automatically create relationship instance for reference fields.

    Example:
        class MyTable(TypedTable):
            reference: OtherTable

    `reference` contains the id of an Other Table row.
     MyTable.relationships should have 'reference' as a relationship, so `MyTable.join('reference')` should work.

     This function will automatically perform this logic (called in db.define):
        to_relationship(MyTable, 'reference', OtherTable) -> Relationship[OtherTable]

    Also works for list:reference (list[OtherTable]) and TypedField[OtherTable].
    """
    if looks_like(field, TypedField):
        # typing.get_args works for list[str] but not for TypedField[role] :(
        if args := t.get_args(field):
            # TypedField[SomeType] -> SomeType
            field = args[0]
        elif hasattr(field, "_type"):
            # TypedField(SomeType) -> SomeType
            field = t.cast(T_Field, field._type)
        else:  # pragma: no cover
            # weird
            return None

    field, optional = extract_type_optional(field)

    try:
        condition = _generate_relationship_condition(cls, key, field)
    except Exception as e:  # pragma: no cover
        warnings.warn("Could not generate Relationship condition", source=e)
        condition = None

    if not condition:  # pragma: no cover
        # something went wrong, not a valid relationship
        warnings.warn(f"Invalid relationship for {cls.__name__}.{key}: {field}")
        return None

    join = "left" if optional or t.get_origin(field) is list else "inner"

    return Relationship(t.cast(type[TypedTable], field), condition, t.cast(JOIN_OPTIONS, join))


# note: these imports exist at the bottom of this file to prevent circular import issues:

from .tables import TypedTable  # noqa: E402



import sqlite3

import pytest

from src.typedal import TypeDAL
from src.typedal.helpers import process_tstring, sql_escape, sql_escape_template, sql_expression


def test_process_tstring_basic(database: TypeDAL):
    """Test the basic f-string functionality example from process_tstring docstring."""

    def fstring_operation(interpolation):
        return str(interpolation.value)

    value = "test"
    template = t"{value = }"
    result = process_tstring(template, fstring_operation)
    assert result == "value = test"


def test_sql_escape_template_security(database: TypeDAL):
    """Test the SQL injection prevention example from sql_escape_template docstring."""
    user_input = "'; DROP TABLE users; --"
    query = t"SELECT * FROM users WHERE name = {user_input}"
    safe_query = sql_escape_template(database, query)

    # The exact escaping format depends on the database adapter, but it should be escaped
    assert "DROP TABLE users" in safe_query  # The dangerous part should still be there
    assert safe_query != f"SELECT * FROM users WHERE name = {user_input}"  # But it should be escaped
    # For most SQL adapters, strings are wrapped in quotes and escaped
    assert "'" in safe_query or '"' in safe_query  # Should have some form of quoting


def test_sql_escape_positional_example(database: TypeDAL):
    """Test the positional arguments example from sql_escape docstring."""
    user_id = 123
    safe_sql = sql_escape(database, "SELECT * FROM users WHERE id = %s", user_id)
    assert safe_sql == "SELECT * FROM users WHERE id = 123"


def test_sql_escape_keyword_example(database: TypeDAL):
    """Test the keyword arguments example from sql_escape docstring."""
    username = "john_doe"
    safe_sql = sql_escape(database, "SELECT * FROM users WHERE name = %(name)s", name=username)
    assert safe_sql == "SELECT * FROM users WHERE name = 'john_doe'"


def test_sql_escape_template_example(database: TypeDAL):
    """Test the Template string example from sql_escape docstring."""
    user_id = 456
    safe_sql = sql_escape(database, t"SELECT * FROM users WHERE id = {user_id}")
    assert safe_sql == "SELECT * FROM users WHERE id = 456"


def test_sql_expression_complex_where(database: TypeDAL):
    """Test the complex WHERE clause example from sql_expression docstring."""
    expr = sql_expression(database, "age > %s AND status = %s", 18, "active", output_type="boolean")

    expected = "age > 18 AND status = 'active'"
    assert str(expr) == expected
    assert expr.type == "boolean"


def test_sql_expression_keyword_extract(database: TypeDAL):
    """Test the keyword arguments EXTRACT example from sql_expression docstring."""
    expr = sql_expression(
        database, "EXTRACT(year FROM %(date_col)s) = %(year)s", date_col="created_at", year=2023, output_type="boolean"
    )

    expected = "EXTRACT(year FROM 'created_at') = 2023"
    assert str(expr) == expected
    assert expr.type == "boolean"


def test_sql_expression_template_age(database: TypeDAL):
    """Test the Template string age example from sql_expression docstring."""
    min_age = 21
    expr = sql_expression(database, t"age >= {min_age}", output_type="boolean")

    expected = "age >= 21"
    assert str(expr) == expected
    assert expr.type == "boolean"


def test_date_expression_similar_to_other_test(database: TypeDAL):
    start_date = "2025-01-01"
    expr1 = database.sql_expression(t"date('now') > {start_date}")
    assert str(expr1) == "date('now') > '2025-01-01'"


def test_executesql_without_tstring(database: TypeDAL):
    bobby_tables = "Robert'); DROP TABLE Students;--"

    database.executesql(f"""
    CREATE TABLE hackable (
        name VARCHAR(100)
    )
    """)

    with pytest.raises(sqlite3.OperationalError):
        database.executesql(f"INSERT INTO hackable(name) VALUES ({bobby_tables})")

    with pytest.raises(sqlite3.OperationalError):
        database.executesql(f"SELECT * FROM hackable where name = {bobby_tables}")


def test_executesql_with_tstring(database: TypeDAL):
    bobby_tables = "Robert'); DROP TABLE Students;--"

    database.executesql(t"""
    CREATE TABLE unhackable (
        name VARCHAR(100)
    )
    """)

    database.executesql(t"INSERT INTO unhackable(name) VALUES ({bobby_tables})")

    rows = database.executesql(t"SELECT * FROM unhackable where name = {bobby_tables}")

    assert len(rows) == 1
    assert rows[0][0] == bobby_tables

    # alternative using magic:
    name = bobby_tables
    rows = database.executesql(t"SELECT * FROM unhackable where {name = }")

    assert len(rows) == 1
    assert rows[0][0] == bobby_tables


def test_sql_expression_314(database: TypeDAL):
    """Main test function that calls all example tests to verify docstring examples."""
    # Call all the docstring example tests
    test_process_tstring_basic(database)
    test_sql_escape_template_security(database)
    test_sql_escape_positional_example(database)
    test_sql_escape_keyword_example(database)
    test_sql_escape_template_example(database)
    test_sql_expression_complex_where(database)
    test_sql_expression_keyword_extract(database)
    test_sql_expression_template_age(database)
    # + the one similar to the non-tstring test:
    test_date_expression_similar_to_other_test(database)
    # executesql with string:
    test_executesql_without_tstring(database)
    test_executesql_with_tstring(database)



import time
from typing import Optional

import pytest

from src.typedal import TypeDAL, TypedTable
from src.typedal.mixins import Mixin, SlugMixin, TimestampsMixin


class AllMixins(TypedTable, SlugMixin, TimestampsMixin, slug_field="name"):
    name: str


class TableWithMixins(TypedTable, SlugMixin, slug_field="name", slug_suffix_length=0):
    name: str
    number: Optional[int]


class TableWithSlugSuffix(TypedTable, SlugMixin, slug_field="name", slug_suffix_length=1):
    name: str
    number: Optional[int]


with pytest.warns(DeprecationWarning):

    class TableWithMixinsWarns(TypedTable, SlugMixin, slug_field="name", slug_suffix=1):
        name: str
        number: Optional[int]


class TableWithTimestamps(TypedTable, TimestampsMixin):
    unrelated: str


def test_invalid_slug_initialization():
    with pytest.raises(ValueError):

        class WithoutSlugField(TypedTable, SlugMixin):  # no slug_field=...
            ...


@pytest.fixture
def db():
    _db = TypeDAL("sqlite:memory")

    _db.define(AllMixins)
    _db.define(TableWithMixins)
    _db.define(TableWithSlugSuffix)
    _db.define(TableWithTimestamps)
    yield _db


def test_order(db):
    assert TableWithMixins.fields == [
        "id",  # id should always be the first (pydal itself already does this);
        "name",  # table own fields should come afterwards,
        "number",  # in the order they are defined;
        "slug",  # mixins should come last.
    ]


def test_slug(db):
    row, error = TableWithMixins.validate_and_insert(name="")
    assert row is None
    assert error

    # without random suffix: duplicates are forbidden

    row, error = TableWithMixins.validate_and_insert(name="Two Words")
    assert error is None

    assert row.name == "Two Words"
    assert row.slug == "two-words"

    assert TableWithMixins.from_slug(row.slug)
    assert TableWithMixins.from_slug("missing") is None

    assert TableWithMixins.from_slug_or_fail(row.slug)

    with pytest.raises(ValueError):
        TableWithMixins.from_slug_or_fail("missing")

    row, error = TableWithMixins.validate_and_insert(name="Two Words")
    assert row is None
    assert error == {
        "name": "This slug is not unique: two-words.",
    }

    # with random suffix: duplicates are fine

    row, error = TableWithSlugSuffix.validate_and_insert(name="Two Words")
    assert error is None

    assert row.name == "Two Words"
    assert row.slug.startswith("two-words-")

    row, error = TableWithSlugSuffix.validate_and_insert(name="Two Words")
    assert error is None

    assert row.name == "Two Words"
    assert row.slug.startswith("two-words")

    # test manual slug:
    manual_slug = "some-other-slug-manually-defined"
    # validate_and_insert will fail because 'slug' is not writable
    row = TableWithMixins.insert(name="Some Name", slug=manual_slug)

    assert row.slug == manual_slug


def test_timestamps(db):
    row = TableWithTimestamps.insert(unrelated="Hi")
    db.commit()
    assert row.updated_at == row.created_at

    time.sleep(1)  # make sure datetimes are not equal

    row.update_record(unrelated="Bye")

    updated_row = TableWithTimestamps(id=row.id)

    assert updated_row.updated_at > updated_row.created_at


def test_reusing(db):
    assert str(AllMixins.created_at) == "all_mixins.created_at"
    assert str(AllMixins.slug) == "all_mixins.slug"
    assert str(AllMixins.name) == "all_mixins.name"

    assert str(TableWithMixins.slug) == "table_with_mixins.slug"
    assert str(TableWithMixins.name) == "table_with_mixins.name"

    assert str(TableWithTimestamps.created_at) == "table_with_timestamps.created_at"
    assert str(TableWithTimestamps.unrelated) == "table_with_timestamps.unrelated"


def test_combining_mixins():
    class FirstMixin(Mixin):
        def __init_subclass__(cls, first: str, **kw):
            super().__init_subclass__(**kw)

            cls.__settings__["first"] = first

        @classmethod
        def one(cls):
            return cls.__settings__["first"] == "first" and cls.__settings__["second"] == "second"

    class SecondMixin(Mixin):
        def __init_subclass__(cls, second: str, **kw):
            super().__init_subclass__(**kw)

            cls.__settings__["second"] = second

        @classmethod
        def two(cls):
            return cls.__settings__["first"] == "first" and cls.__settings__["second"] == "second"

    class Combined(TypedTable, FirstMixin, SecondMixin, first="first", second="second"): ...

    assert Combined.one()
    assert Combined.two()

    class CombinedDifferentOrder(TypedTable, SecondMixin, FirstMixin, first="first", second="second"): ...

    assert CombinedDifferentOrder.one()
    assert CombinedDifferentOrder.two()



"""
This file contains available Field types.
"""

from __future__ import annotations

import ast
import contextlib
import datetime as dt
import decimal
import types
import typing as t
import uuid

import pydal
from pydal.helpers.classes import SQLCustomType
from pydal.objects import Table

from .core import TypeDAL
from .types import (
    Expression,
    Field,
    FieldSettings,
    Query,
    T_annotation,
    T_MetaInstance,
    T_subclass,
    T_Value,
    Validator,
)

if t.TYPE_CHECKING:
    # will be imported for real later:
    from .tables import TypedTable


## general


class TypedField(Expression, t.Generic[T_Value]):  # pragma: no cover
    """
    Typed version of pydal.Field, which will be converted to a normal Field in the background.
    """

    # will be set by .bind on db.define
    name = ""
    _db: t.Optional[pydal.DAL] = None
    _rname: t.Optional[str] = None
    _table: t.Optional[Table] = None
    _field: t.Optional[Field] = None

    _type: T_annotation
    kwargs: t.Any

    requires: Validator | t.Iterable[Validator]

    # NOTE: for the logic of converting a TypedField into a pydal Field, see TypeDAL._to_field

    def __init__(
        self,
        _type: t.Type[T_Value] | types.UnionType = str,  # type: ignore
        /,
        **settings: t.Unpack[FieldSettings],
    ) -> None:
        """
        Typed version of pydal.Field, which will be converted to a normal Field in the background.

        Provide the Python type for this field as the first positional argument
        and any other settings to Field() as keyword parameters.
        """
        self._type = _type
        self.kwargs = settings
        # super().__init__()

    @t.overload
    def __get__(self, instance: T_MetaInstance, owner: t.Type[T_MetaInstance]) -> T_Value:  # pragma: no cover
        """
        row.field -> (actual data).
        """

    @t.overload
    def __get__(self, instance: None, owner: "t.Type[TypedTable]") -> "TypedField[T_Value]":  # pragma: no cover
        """
        Table.field -> Field.
        """

    def __get__(
        self,
        instance: T_MetaInstance | None,
        owner: t.Type[T_MetaInstance],
    ) -> t.Union[T_Value, "TypedField[T_Value]"]:
        """
        Since this class is a Descriptor field, \
            it returns something else depending on if it's called on a class or instance.

        (this is mostly for mypy/typing)
        """
        if instance:
            # this is only reached in a very specific case:
            # an instance of the object was created with a specific set of fields selected (excluding the current one)
            # in that case, no value was stored in the owner -> return None (since the field was not selected)
            return t.cast(T_Value, None)  # cast as T_Value so mypy understands it for selected fields
        else:
            # getting as class -> return actual field so pydal understands it when using in query etc.
            return t.cast(TypedField[T_Value], self._field)  # pretend it's still typed for IDE support

    def __str__(self) -> str:
        """
        String representation of a Typed Field.

        If `type` is set explicitly (e.g. TypedField(str, type="text")), that type is used: `TypedField.text`,
        otherwise the type annotation is used (e.g. TypedField(str) -> TypedField.str)
        """
        return str(self._field) if self._field else ""

    def __repr__(self) -> str:
        """
        More detailed string representation of a Typed Field.

        Uses __str__ and adds the provided extra options (kwargs) in the representation.
        """
        string_value = self.__str__()

        if "type" in self.kwargs:
            # manual type in kwargs supplied
            typename = self.kwargs["type"]
        elif issubclass(type, type(self._type)):
            # normal type, str.__name__ = 'str'
            typename = getattr(self._type, "__name__", str(self._type))
        elif t_args := t.get_args(self._type):
            # list[str] -> 'str'
            typename = t_args[0].__name__
        else:  # pragma: no cover
            # fallback - something else, may not even happen, I'm not sure
            typename = self._type

        string_value = f"TypedField[{typename}].{string_value}" if string_value else f"TypedField[{typename}]"

        kw = self.kwargs.copy()
        kw.pop("type", None)
        return f"<{string_value} with options {kw}>"

    def _to_field(self, extra_kwargs: t.MutableMapping[str, t.Any], builder: TableDefinitionBuilder) -> t.Optional[str]:
        """
        Convert a Typed Field instance to a pydal.Field.

        Actual logic in TypeDAL._to_field but this function creates the pydal type name and updates the kwarg settings.
        """
        other_kwargs = self.kwargs.copy()
        extra_kwargs.update(other_kwargs)  # <- modifies and overwrites the default kwargs with user-specified ones
        return extra_kwargs.pop("type", False) or builder.annotation_to_pydal_fieldtype(
            self._type,
            extra_kwargs,
        )

    def bind(self, field: pydal.objects.Field, table: pydal.objects.Table) -> None:
        """
        Bind the right db/table/field info to this class, so queries can be made using `Class.field == ...`.
        """
        self._table = table
        self._field = field

    def __getattr__(self, key: str) -> t.Any:
        """
        If the regular getattribute does not work, try to get info from the related Field.
        """
        with contextlib.suppress(AttributeError):
            return super().__getattribute__(key)

        # try on actual field:
        return getattr(self._field, key)

    def __eq__(self, other: t.Any) -> Query:
        """
        Performing == on a Field will result in a Query.
        """
        return t.cast(Query, self._field == other)

    def __ne__(self, other: t.Any) -> Query:
        """
        Performing != on a Field will result in a Query.
        """
        return t.cast(Query, self._field != other)

    def __gt__(self, other: t.Any) -> Query:
        """
        Performing > on a Field will result in a Query.
        """
        return t.cast(Query, self._field > other)

    def __lt__(self, other: t.Any) -> Query:
        """
        Performing < on a Field will result in a Query.
        """
        return t.cast(Query, self._field < other)

    def __ge__(self, other: t.Any) -> Query:
        """
        Performing >= on a Field will result in a Query.
        """
        return t.cast(Query, self._field >= other)

    def __le__(self, other: t.Any) -> Query:
        """
        Performing <= on a Field will result in a Query.
        """
        return t.cast(Query, self._field <= other)

    def __hash__(self) -> int:
        """
        Shadow Field.__hash__.
        """
        return hash(self._field)

    def __invert__(self) -> Expression:
        """
        Performing ~ on a Field will result in an Expression.
        """
        if not self._field:  # pragma: no cover
            raise ValueError("Unbound Field can not be inverted!")

        return t.cast(Expression, ~self._field)

    def lower(self) -> Expression:
        """
        For string-fields: compare lowercased values.
        """
        if not self._field:  # pragma: no cover
            raise ValueError("Unbound Field can not be lowered!")

        return t.cast(Expression, self._field.lower())


def is_typed_field(cls: t.Any) -> t.TypeGuard["TypedField[t.Any]"]:
    """
    Is `cls` an instance or subclass of TypedField?

    Deprecated
    """
    return isinstance(cls, TypedField) or (
        isinstance(t.get_origin(cls), type) and issubclass(t.get_origin(cls), TypedField)
    )


## specific
def StringField(**kw: t.Unpack[FieldSettings]) -> TypedField[str]:
    """
    Pydal type is string, Python type is str.
    """
    kw["type"] = "string"
    return TypedField(str, **kw)


String = StringField


def TextField(**kw: t.Unpack[FieldSettings]) -> TypedField[str]:
    """
    Pydal type is text, Python type is str.
    """
    kw["type"] = "text"
    return TypedField(str, **kw)


Text = TextField


def BlobField(**kw: t.Unpack[FieldSettings]) -> TypedField[bytes]:
    """
    Pydal type is blob, Python type is bytes.
    """
    kw["type"] = "blob"
    return TypedField(bytes, **kw)


Blob = BlobField


def BooleanField(**kw: t.Unpack[FieldSettings]) -> TypedField[bool]:
    """
    Pydal type is boolean, Python type is bool.
    """
    kw["type"] = "boolean"
    return TypedField(bool, **kw)


Boolean = BooleanField


def IntegerField(**kw: t.Unpack[FieldSettings]) -> TypedField[int]:
    """
    Pydal type is integer, Python type is int.
    """
    kw["type"] = "integer"
    return TypedField(int, **kw)


Integer = IntegerField


def DoubleField(**kw: t.Unpack[FieldSettings]) -> TypedField[float]:
    """
    Pydal type is double, Python type is float.
    """
    kw["type"] = "double"
    return TypedField(float, **kw)


Double = DoubleField


def DecimalField(n: int, m: int, **kw: t.Unpack[FieldSettings]) -> TypedField[decimal.Decimal]:
    """
    Pydal type is decimal, Python type is Decimal.
    """
    kw["type"] = f"decimal({n}, {m})"
    return TypedField(decimal.Decimal, **kw)


Decimal = DecimalField


def DateField(**kw: t.Unpack[FieldSettings]) -> TypedField[dt.date]:
    """
    Pydal type is date, Python type is datetime.date.
    """
    kw["type"] = "date"
    return TypedField(dt.date, **kw)


Date = DateField


def TimeField(**kw: t.Unpack[FieldSettings]) -> TypedField[dt.time]:
    """
    Pydal type is time, Python type is datetime.time.
    """
    kw["type"] = "time"
    return TypedField(dt.time, **kw)


Time = TimeField


def DatetimeField(**kw: t.Unpack[FieldSettings]) -> TypedField[dt.datetime]:
    """
    Pydal type is datetime, Python type is datetime.datetime.
    """
    kw["type"] = "datetime"
    return TypedField(dt.datetime, **kw)


Datetime = DatetimeField


def PasswordField(**kw: t.Unpack[FieldSettings]) -> TypedField[str]:
    """
    Pydal type is password, Python type is str.
    """
    kw["type"] = "password"
    return TypedField(str, **kw)


Password = PasswordField


def UploadField(**kw: t.Unpack[FieldSettings]) -> TypedField[str]:
    """
    Pydal type is upload, Python type is str.
    """
    kw["type"] = "upload"
    return TypedField(str, **kw)


Upload = UploadField


def ReferenceField(
    other_table: str | t.Type[TypedTable] | TypedTable | Table | T_subclass,
    **kw: t.Unpack[FieldSettings],
) -> TypedField[int]:
    """
    Pydal type is reference, Python type is int (id).
    """
    if isinstance(other_table, str):
        kw["type"] = f"reference {other_table}"
    elif isinstance(other_table, Table):
        # db.table
        kw["type"] = f"reference {other_table._tablename}"
    elif isinstance(other_table, type) and issubclass(other_table, TypedTable):
        # SomeTable
        snakename = TypeDAL.to_snake(other_table.__name__)
        kw["type"] = f"reference {snakename}"
    else:
        raise ValueError(f"Don't know what to do with {type(other_table)}")

    return TypedField(int, **kw)


Reference = ReferenceField


def ListStringField(**kw: t.Unpack[FieldSettings]) -> TypedField[list[str]]:
    """
    Pydal type is list:string, Python type is list of str.
    """
    kw["type"] = "list:string"
    return TypedField(list[str], **kw)


ListString = ListStringField


def ListIntegerField(**kw: t.Unpack[FieldSettings]) -> TypedField[list[int]]:
    """
    Pydal type is list:integer, Python type is list of int.
    """
    kw["type"] = "list:integer"
    return TypedField(list[int], **kw)


ListInteger = ListIntegerField


def ListReferenceField(other_table: str, **kw: t.Unpack[FieldSettings]) -> TypedField[list[int]]:
    """
    Pydal type is list:reference, Python type is list of int (id).
    """
    kw["type"] = f"list:reference {other_table}"
    return TypedField(list[int], **kw)


ListReference = ListReferenceField


def JSONField(**kw: t.Unpack[FieldSettings]) -> TypedField[object]:
    """
    Pydal type is json, Python type is object (can be anything JSON-encodable).
    """
    kw["type"] = "json"
    return TypedField(object, **kw)


def BigintField(**kw: t.Unpack[FieldSettings]) -> TypedField[int]:
    """
    Pydal type is bigint, Python type is int.
    """
    kw["type"] = "bigint"
    return TypedField(int, **kw)


Bigint = BigintField

## Custom:

NativeTimestampField = SQLCustomType(
    type="datetime",
    native="timestamp",
    encoder=lambda x: f"'{x}'",  # extra quotes
    # decoder=lambda x: x, # already parsed into datetime
)


def TimestampField(**kw: t.Unpack[FieldSettings]) -> TypedField[dt.datetime]:
    """
    Database type is timestamp, Python type is datetime.

    Advantage over the regular datetime type is that
    a timestamp has millisecond precision (2024-10-11 20:18:24.505194)
    whereas a regular datetime only has precision up to the second (2024-10-11 20:18:24)
    """
    kw["type"] = NativeTimestampField
    return TypedField(
        dt.datetime,
        **kw,
    )


def safe_decode_native_point(value: str | None) -> tuple[float, ...]:
    """
    Safely decode a string into a tuple of floats.

    The function attempts to parse the input string using `ast.literal_eval`.
    If the parsing is successful, the function casts the parsed value to a tuple of floats and returns it.
    Otherwise, the function returns an empty tuple.

    Args:
        value: The string to decode.

    Returns:
        A tuple of floats.
    """
    if not value:
        return ()

    try:
        parsed = ast.literal_eval(value)
        return t.cast(tuple[float, ...], parsed)
    except ValueError:  # pragma: no cover
        # should not happen when inserted with `safe_encode_native_point` but you never know
        return ()


def safe_encode_native_point(value: tuple[str, str] | tuple[float, float] | str) -> str:
    """

    Safe encodes a point value.

    The function takes a point value as input.
    It can be a string in the format "x,y" or a tuple of two numbers.
    The function converts the string to a tuple if necessary, validates the tuple,
    and formats it into the expected string format.

    Args:
        value: The point value to be encoded.

    Returns:
        The encoded point value as a string in the format "x,y".

    Raises:
        ValueError: If the input value is not a valid point.
    """
    if not value:
        return ""

    # Convert string to tuple if needed
    if isinstance(value, str):
        value = value.strip("() ")
        if not value:
            return ""
        value_tup = tuple(float(x.strip()) for x in value.split(","))
    else:
        value_tup = value  # type: ignore

    # Validate and format
    if len(value_tup) != 2:
        raise ValueError("Point must have exactly 2 coordinates")

    x, y = value_tup
    return f"({x},{y})"


NativePointField = SQLCustomType(
    type="string",
    native="point",
    encoder=safe_encode_native_point,
    decoder=safe_decode_native_point,
)


def PointField(**kw: t.Unpack[FieldSettings]) -> TypedField[tuple[float, float]]:
    """
    Database type is point, Python type is tuple[float, float].
    """
    kw["type"] = NativePointField
    return TypedField(tuple[float, float], **kw)


NativeUUIDField = SQLCustomType(
    type="string",
    native="uuid",
    encoder=str,
    decoder=lambda value: uuid.UUID(value) if value else None,
)


def UUIDField(**kw: t.Unpack[FieldSettings]) -> TypedField[uuid.UUID]:
    """
    Database type is uuid, Python type is UUID.
    """
    kw["type"] = NativeUUIDField
    return TypedField(uuid.UUID, **kw)


# note: import at the end to prevent circular imports:
from .define import TableDefinitionBuilder  # noqa: E402
from .tables import TypedTable  # noqa: E402



import pytest

from src.typedal import TypeDAL, TypedTable
from src.typedal.caching import calculate_stats, humanize_bytes, row_stats, table_stats


class SomeCachedTable(TypedTable):
    key: str
    value: int


@pytest.fixture
def database():
    db = TypeDAL("sqlite:memory")

    db.define(SomeCachedTable)

    SomeCachedTable.bulk_insert(
        [
            {"key": "first", "value": 1},
            {"key": "second", "value": 2},
            {"key": "third", "value": 3},
            {"key": "fourth", "value": 4},
        ]
    )

    yield db


def test_humanize_bytes():
    assert humanize_bytes(0) == humanize_bytes(None) == "0"

    assert "MB" in humanize_bytes(10000000)


def test_stats(database):
    assert len(SomeCachedTable.cache().collect()) == 4
    assert len(SomeCachedTable.cache().collect()) == 4

    generic = calculate_stats(database)
    assert generic["total"]["dependencies"] == generic["valid"]["dependencies"] == 4
    assert generic["expired"]["dependencies"] == 0

    table = table_stats(database, "some_cached_table")
    assert table["total"] == table["valid"]

    row = row_stats(database, "some_cached_table", "1")
    assert row["total"]["Dependent Cache Entries"] == row["valid"]["Dependent Cache Entries"] == 1



"""
Contains base functionality related to Rows (raw result of a database query).
"""

from __future__ import annotations

import csv
import json
import typing as t

import pydal.objects

from .core import TypeDAL
from .helpers import mktable
from .query_builder import QueryBuilder
from .serializers import as_json
from .tables import TypedTable
from .types import (
    AnyDict,
    Field,
    Metadata,
    PaginateDict,
    Pagination,
    Query,
    Row,
    Rows,
    T,
    T_MetaInstance,
)


class TypedRows(t.Collection[T_MetaInstance], Rows):
    """
    Slighly enhaned and typed functionality on top of pydal Rows (the result of a select).
    """

    records: dict[int, T_MetaInstance]
    # _rows: Rows
    model: t.Type[T_MetaInstance]
    metadata: Metadata

    # pseudo-properties: actually stored in _rows
    db: TypeDAL
    colnames: list[str]
    fields: list[Field]
    colnames_fields: list[Field]
    response: list[tuple[t.Any, ...]]

    def __init__(
        self,
        rows: Rows,
        model: t.Type[T_MetaInstance],
        records: dict[int, T_MetaInstance] = None,
        metadata: Metadata = None,
        raw: dict[int, list[Row]] = None,
    ) -> None:
        """
        Should not be called manually!

        Normally, the `records` from an existing `Rows` object are used
            but these can be overwritten with a `records` dict.
        `metadata` can be t.Any (un)structured data
        `model` is a Typed Table class
        """

        def _get_id(row: Row) -> int:
            """
            Try to find the id field in a row.

            If _extra exists, the row changes:
            <Row {'test_relationship': {'id': 1}, '_extra': {'COUNT("test_relationship"."querytable")': 8}}>
            """
            if idx := getattr(row, "id", None):
                return t.cast(int, idx)
            elif main := getattr(row, str(model), None):
                return t.cast(int, main.id)
            else:  # pragma: no cover
                raise NotImplementedError(f"`id` could not be found for {row}")

        records = records or {_get_id(row): model(row) for row in rows}
        raw = raw or {}

        for idx, entity in records.items():
            entity._rows = tuple(raw.get(idx, []))

        super().__init__(rows.db, records, rows.colnames, rows.compact, rows.response, rows.fields)
        self.model = model
        self.metadata = metadata or {}
        self.colnames = rows.colnames

    def __len__(self) -> int:
        """
        Return the count of rows.
        """
        return len(self.records)

    def __iter__(self) -> t.Iterator[T_MetaInstance]:
        """
        Loop through the rows.
        """
        yield from self.records.values()

    def __contains__(self, ind: t.Any) -> bool:
        """
        Check if an id exists in this result set.
        """
        return ind in self.records

    def first(self) -> T_MetaInstance | None:
        """
        Get the row with the lowest id.
        """
        if not self.records:
            return None

        return next(iter(self))

    def last(self) -> T_MetaInstance | None:
        """
        Get the row with the highest id.
        """
        if not self.records:
            return None

        max_id = max(self.records.keys())
        return self[max_id]

    def find(
        self,
        f: t.Callable[[T_MetaInstance], Query],
        limitby: tuple[int, int] = None,
    ) -> "TypedRows[T_MetaInstance]":
        """
        Returns a new Rows object, a subset of the original object, filtered by the function `f`.
        """
        if not self.records:
            return self.__class__(self, self.model, {})

        records = {}
        if limitby:
            _min, _max = limitby
        else:
            _min, _max = 0, len(self)
        count = 0
        for i, row in self.records.items():
            if f(row):
                if _min <= count:
                    records[i] = row
                count += 1
                if count == _max:
                    break

        return self.__class__(self, self.model, records)

    def exclude(self, f: t.Callable[[T_MetaInstance], Query]) -> "TypedRows[T_MetaInstance]":
        """
        Removes elements from the calling Rows object, filtered by the function `f`, \
            and returns a new Rows object containing the removed elements.
        """
        if not self.records:
            return self.__class__(self, self.model, {})
        removed = {}
        to_remove = []
        for i in self.records:
            row = self[i]
            if f(row):
                removed[i] = self.records[i]
                to_remove.append(i)

        [self.records.pop(i) for i in to_remove]

        return self.__class__(
            self,
            self.model,
            removed,
        )

    def sort(self, f: t.Callable[[T_MetaInstance], t.Any], reverse: bool = False) -> list[T_MetaInstance]:
        """
        Returns a list of sorted elements (not sorted in place).
        """
        return [r for (r, s) in sorted(zip(self.records.values(), self), key=lambda r: f(r[1]), reverse=reverse)]

    def __str__(self) -> str:
        """
        Simple string representation.
        """
        return f"<TypedRows with {len(self)} records>"

    def __repr__(self) -> str:
        """
        Print a table on repr().
        """
        data = self.as_dict()
        try:
            headers = list(next(iter(data.values())).keys())
        except StopIteration:
            headers = []

        return mktable(data, headers)

    def group_by_value(
        self,
        *fields: "str | Field | TypedField[T]",
        one_result: bool = False,
        **kwargs: t.Any,
    ) -> dict[T, list[T_MetaInstance]]:
        """
        Group the rows by a specific field (which will be the dict key).
        """
        kwargs["one_result"] = one_result
        result = super().group_by_value(*fields, **kwargs)
        return t.cast(dict[T, list[T_MetaInstance]], result)

    def as_csv(self) -> str:
        """
        Dump the data to csv.
        """
        return t.cast(str, super().as_csv())

    def as_dict(
        self,
        key: str | Field | None = None,
        compact: bool = False,
        storage_to_dict: bool = False,
        datetime_to_str: bool = False,
        custom_types: list[type] | None = None,
    ) -> dict[int, AnyDict]:
        """
        Get the data in a dict of dicts.
        """
        if any([key, compact, storage_to_dict, datetime_to_str, custom_types]):
            # functionality not guaranteed
            if isinstance(key, Field):
                key = key.name

            return t.cast(
                dict[int, AnyDict],
                super().as_dict(
                    key or "id",
                    compact,
                    storage_to_dict,
                    datetime_to_str,
                    custom_types,
                ),
            )

        return {k: v.as_dict() for k, v in self.records.items()}

    def as_json(
        self, default: t.Callable[[t.Any], t.Any] = None, indent: t.Optional[int] = None, **kwargs: t.Any
    ) -> str:
        """
        Turn the data into a dict and then dump to JSON.
        """
        data = self.as_list()

        return as_json.encode(data, default=default, indent=indent, **kwargs)

    def json(self, default: t.Callable[[t.Any], t.Any] = None, indent: t.Optional[int] = None, **kwargs: t.Any) -> str:
        """
        Turn the data into a dict and then dump to JSON.
        """
        return self.as_json(default=default, indent=indent, **kwargs)

    def as_list(
        self,
        compact: bool = False,
        storage_to_dict: bool = False,
        datetime_to_str: bool = False,
        custom_types: list[type] = None,
    ) -> list[AnyDict]:
        """
        Get the data in a list of dicts.
        """
        if any([compact, storage_to_dict, datetime_to_str, custom_types]):
            return t.cast(list[AnyDict], super().as_list(compact, storage_to_dict, datetime_to_str, custom_types))

        return [_.as_dict() for _ in self.records.values()]

    def __getitem__(self, item: int) -> T_MetaInstance:
        """
        You can get a specific row by ID from a typedrows by using rows[idx] notation.

        Since pydal's implementation differs (they expect a list instead of a dict with id keys),
        using rows[0] will return the first row, regardless of its id.
        """
        try:
            return self.records[item]
        except KeyError as e:
            if item == 0 and (row := self.first()):
                # special case: pydal internals think Rows.records is a list, not a dict
                return row

            raise e

    def get(self, item: int) -> t.Optional[T_MetaInstance]:
        """
        Get a row by ID, or receive None if it isn't in this result set.
        """
        return self.records.get(item)

    def update(self, **new_values: t.Any) -> bool:
        """
        Update the current rows in the database with new_values.
        """
        # cast to make mypy understand .id is a TypedField and not an int!
        table = t.cast(t.Type[TypedTable], self.model._ensure_table_defined())

        ids = set(self.column("id"))
        query = table.id.belongs(ids)
        return bool(self.db(query).update(**new_values))

    def delete(self) -> bool:
        """
        Delete the currently selected rows from the database.
        """
        # cast to make mypy understand .id is a TypedField and not an int!
        table = t.cast(t.Type[TypedTable], self.model._ensure_table_defined())

        ids = set(self.column("id"))
        query = table.id.belongs(ids)
        return bool(self.db(query).delete())

    def join(
        self,
        field: "Field | TypedField[t.Any]",
        name: str = None,
        constraint: Query = None,
        fields: list[str | Field] = None,
        orderby: t.Optional[str | Field] = None,
    ) -> T_MetaInstance:
        """
        This can be used to JOIN with some relationships after the initial select.

        Using the querybuilder's .join() method is prefered!
        """
        result = super().join(field, name, constraint, fields or [], orderby)
        return t.cast(T_MetaInstance, result)

    def export_to_csv_file(
        self,
        ofile: t.TextIO,
        null: t.Any = "<NULL>",
        delimiter: str = ",",
        quotechar: str = '"',
        quoting: int = csv.QUOTE_MINIMAL,
        represent: bool = False,
        colnames: list[str] = None,
        write_colnames: bool = True,
        *args: t.Any,
        **kwargs: t.Any,
    ) -> None:
        """
        Shadow export_to_csv_file from Rows, but with typing.

        See http://web2py.com/books/default/chapter/29/06/the-database-abstraction-layer?search=export_to_csv_file#Exporting-and-importing-data
        """
        super().export_to_csv_file(
            ofile,
            null,
            *args,
            delimiter=delimiter,
            quotechar=quotechar,
            quoting=quoting,
            represent=represent,
            colnames=colnames or self.colnames,
            write_colnames=write_colnames,
            **kwargs,
        )

    @classmethod
    def from_rows(
        cls,
        rows: Rows,
        model: t.Type[T_MetaInstance],
        metadata: Metadata = None,
    ) -> "TypedRows[T_MetaInstance]":
        """
        Internal method to convert a Rows object to a TypedRows.
        """
        return cls(rows, model, metadata=metadata)

    def __getstate__(self) -> AnyDict:
        """
        Used by dill to dump to bytes (exclude db connection etc).
        """
        return {
            "metadata": json.dumps(self.metadata, default=str),
            "records": self.records,
            "model": str(self.model._table),
            "colnames": self.colnames,
        }

    def __setstate__(self, state: AnyDict) -> None:
        """
        Used by dill when loading from a bytestring.
        """
        state["metadata"] = json.loads(state["metadata"])
        self.__dict__.update(state)
        # db etc. set after undill by caching.py

    def render(
        self,
        i: int | None = None,
        fields: list[Field] | None = None,
    ) -> t.Generator[T_MetaInstance, None, None]:
        """
        Takes an index and returns a copy of the indexed row with values \
            transformed via the "represent" attributes of the associated fields.

        Args:
            i: index. If not specified, a generator is returned for iteration
                over all the rows.
            fields: a list of fields to transform (if None, all fields with
                "represent" attributes will be transformed)
        """
        if i is None:
            # difference: uses .keys() instead of index
            return (self.render(i, fields=fields) for i in self.records)

        if not self.db.has_representer("rows_render"):  # pragma: no cover
            raise RuntimeError(
                "Rows.render() needs a `rows_render` representer in DAL instance",
            )

        row = self.records[i]
        return row.render(fields, compact=self.compact)


class PaginatedRows(TypedRows[T_MetaInstance]):
    """
    Extension on top of rows that is used when calling .paginate() instead of .collect().
    """

    _query_builder: QueryBuilder[T_MetaInstance]

    @property
    def data(self) -> list[T_MetaInstance]:
        """
        Get the underlying data.
        """
        return list(self.records.values())

    @property
    def pagination(self) -> Pagination:
        """
        Get all page info.
        """
        pagination_data = self.metadata["pagination"]

        has_next_page = pagination_data["current_page"] < pagination_data["max_page"]
        has_prev_page = pagination_data["current_page"] > 1
        return {
            "total_items": pagination_data["rows"],
            "current_page": pagination_data["current_page"],
            "per_page": pagination_data["limit"],
            "total_pages": pagination_data["max_page"],
            "has_next_page": has_next_page,
            "has_prev_page": has_prev_page,
            "next_page": pagination_data["current_page"] + 1 if has_next_page else None,
            "prev_page": pagination_data["current_page"] - 1 if has_prev_page else None,
        }

    def next(self) -> t.Self:
        """
        Get the next page.
        """
        data = self.metadata["pagination"]
        if data["current_page"] >= data["max_page"]:
            raise StopIteration("Final Page")

        return self._query_builder.paginate(limit=data["limit"], page=data["current_page"] + 1)

    def previous(self) -> t.Self:
        """
        Get the previous page.
        """
        data = self.metadata["pagination"]
        if data["current_page"] <= 1:
            raise StopIteration("First Page")

        return self._query_builder.paginate(limit=data["limit"], page=data["current_page"] - 1)

    def as_dict(self, *_: t.Any, **__: t.Any) -> PaginateDict:  # type: ignore
        """
        Convert to a dictionary with pagination info and original data.

        All arguments are ignored!
        """
        return {"data": super().as_dict(), "pagination": self.pagination}


class TypedSet(pydal.objects.Set):  # pragma: no cover
    """
    Used to make pydal Set more typed.

    This class is not actually used, only 'cast' by TypeDAL.__call__
    """

    def count(self, distinct: t.Optional[bool] = None, cache: AnyDict = None) -> int:
        """
        Count returns an int.
        """
        result = super().count(distinct, cache)
        return t.cast(int, result)

    def select(self, *fields: t.Any, **attributes: t.Any) -> TypedRows[T_MetaInstance]:
        """
        Select returns a TypedRows of a user defined table.

        Example:
            result: TypedRows[MyTable] = db(MyTable.id > 0).select()

            for row in result:
                reveal_type(row)  # MyTable
        """
        rows = super().select(*fields, **attributes)
        return t.cast(TypedRows[T_MetaInstance], rows)


# note: these imports exist at the bottom of this file to prevent circular import issues:

from .fields import TypedField  # noqa: E402



"""
Stuff to make mypy happy.
"""

# ---------------------------------------------------------------------------
# Imports
# ---------------------------------------------------------------------------

# Standard library
import datetime as dt
import types
import typing as t

import pydal.objects

# Third-party
from pydal.adapters.base import BaseAdapter
from pydal.helpers.classes import OpRow as _OpRow
from pydal.helpers.classes import Reference as _Reference
from pydal.helpers.classes import SQLCustomType
from pydal.objects import Expression as _Expression
from pydal.objects import Field as _Field
from pydal.objects import Query as _Query
from pydal.objects import Row as _Row
from pydal.objects import Rows as _Rows
from pydal.objects import Set as _Set
from pydal.objects import Table as _Table
from pydal.validators import Validator as _Validator

try:
    from string.templatelib import Template
except ImportError:
    Template: t.TypeAlias = str  # type: ignore

# Internal references
if t.TYPE_CHECKING:
    from .fields import TypedField
    from .tables import TypedTable

# ---------------------------------------------------------------------------
# Aliases
# ---------------------------------------------------------------------------

AnyCallable: t.TypeAlias = t.Callable[..., t.Any]
AnyDict: t.TypeAlias = dict[str, t.Any]


# ---------------------------------------------------------------------------
# Protocols
# ---------------------------------------------------------------------------


class TableProtocol(t.Protocol):  # pragma: no cover
    """Protocol to make mypy happy for Tables."""

    id: "TypedField[int]"

    def __getitem__(self, item: str) -> "Field":
        """
        Tables have table[field] syntax.
        """


class CacheFn(t.Protocol):
    """
    The cache model (e.g. cache.ram) accepts these parameters (all filled by default).
    """

    def __call__(
        self: BaseAdapter,
        sql: str = "",
        fields: t.Iterable[str] = (),
        attributes: t.Iterable[str] = (),
        colnames: t.Iterable[str] = (),
    ) -> "Rows":
        """Signature for calling this object."""


class FileSystemLike(t.Protocol):  # pragma: no cover
    """Protocol for any class that has an 'open' function (e.g. OSFS)."""

    def open(self, file: str, mode: str = "r") -> t.IO[t.Any]:
        """We assume every object with an open function this shape, is basically a file."""


# ---------------------------------------------------------------------------
# pydal Wrappers (to help mypy understand these classes)
# ---------------------------------------------------------------------------


class Query(_Query):
    """Pydal Query object. Makes mypy happy."""


class Expression(_Expression):
    """Pydal Expression object. Make mypy happy."""


class Set(_Set):
    """Pydal Set object. Make mypy happy."""


if t.TYPE_CHECKING:

    class OpRow:
        """
        Pydal OpRow object for typing (otherwise mypy thinks it's Any).
        """

        def __getitem__(self, item: str) -> t.Any:
            """row.item syntax."""

        def __setitem__(self, key: str, value: t.Any) -> None:
            """row.item = key syntax."""

        # more methods could be added

else:

    class OpRow(_OpRow):
        """Runtime OpRow, using pydal's version."""


class Reference(_Reference):
    """Pydal Reference object. Make mypy happy."""


class Field(_Field):
    """Pydal Field object. Make mypy happy."""


class Rows(_Rows):
    """Pydal Rows object. Make mypy happy."""

    def column(self, column: t.Any = None) -> list[t.Any]:
        """
        Get a list of all values in a specific column.

        Example:
            rows.column('name') -> ['Name 1', 'Name 2', ...]
        """
        return [r[str(column) if column else self.colnames[0]] for r in self]


class Row(_Row):
    """Pydal Row object. Make mypy happy."""


class Validator(_Validator):
    """Pydal Validator object. Make mypy happy."""


class Table(_Table, TableProtocol):
    """Table with protocol support. Make mypy happy."""


# ---------------------------------------------------------------------------
# Utility Types
# ---------------------------------------------------------------------------


class _Types:
    """Internal type storage for stuff mypy otherwise won't understand."""

    NONETYPE = type(None)


# ---------------------------------------------------------------------------
# TypedDicts
# ---------------------------------------------------------------------------


class Pagination(t.TypedDict):
    """Pagination key of a paginate dict has these items."""

    total_items: int
    current_page: int
    per_page: int
    total_pages: int
    has_next_page: bool
    has_prev_page: bool
    next_page: t.Optional[int]
    prev_page: t.Optional[int]


class PaginateDict(t.TypedDict):
    """Result of PaginatedRows.as_dict()."""

    data: dict[int, AnyDict]
    pagination: Pagination


CacheStatus = t.Literal["fresh", "cached"]


class CacheMetadata(t.TypedDict):
    """Used by query builder metadata in the 'cache' key."""

    enabled: bool
    depends_on: list[t.Any]
    key: t.NotRequired[str | None]
    status: t.NotRequired[CacheStatus | None]
    expires_at: t.NotRequired[dt.datetime | None]
    cached_at: t.NotRequired[dt.datetime | None]


class PaginationMetadata(t.TypedDict):
    """Used by query builder metadata in the 'pagination' key."""

    limit: int
    current_page: int
    max_page: int
    rows: int
    min_max: tuple[int, int]


class SelectKwargs(t.TypedDict, total=False):
    """Possible keyword arguments for .select()."""

    join: t.Optional[list[Expression]]
    left: t.Optional[list[Expression]]
    orderby: "OrderBy | t.Iterable[OrderBy] | None"
    groupby: "GroupBy | t.Iterable[GroupBy] | None"
    having: "Having | None"
    limitby: t.Optional[tuple[int, int]]
    distinct: bool | Field | Expression
    orderby_on_limitby: bool
    cacheable: bool
    cache: "CacheTuple"


class Metadata(t.TypedDict):
    """Loosely structured metadata used by Query Builder."""

    cache: t.NotRequired[CacheMetadata]
    pagination: t.NotRequired[PaginationMetadata]
    query: t.NotRequired[Query | str | None]
    ids: t.NotRequired[str]
    final_query: t.NotRequired[Query | str | None]
    final_args: t.NotRequired[list[t.Any]]
    final_kwargs: t.NotRequired[SelectKwargs]
    select_duration: t.NotRequired[float]
    relationships: t.NotRequired[set[str]]
    sql: t.NotRequired[str]


class FieldSettings(t.TypedDict, total=False):
    """
    The supported keyword arguments for `pydal.Field()`.

    Other arguments can be passed.
    """

    type: str | type | SQLCustomType
    length: int
    default: t.Any
    required: bool
    requires: list[AnyCallable | t.Any | Validator] | Validator | AnyCallable
    ondelete: str
    onupdate: str
    notnull: bool
    unique: bool
    uploadfield: bool | str
    widget: AnyCallable
    label: str
    comment: str
    writable: bool
    readable: bool
    searchable: bool
    listable: bool
    regex: str
    options: list[t.Any] | AnyCallable
    update: t.Any
    authorize: AnyCallable
    autodelete: bool
    represent: AnyCallable
    uploadfolder: str
    uploadseparate: bool
    uploadfs: FileSystemLike
    compute: AnyCallable
    custom_store: AnyCallable
    custom_retrieve: AnyCallable
    custom_retrieve_file_properties: AnyCallable
    custom_delete: AnyCallable
    filter_in: AnyCallable
    filter_out: AnyCallable
    custom_qualifier: t.Any
    map_none: t.Any
    rname: str


# ---------------------------------------------------------------------------
# Generics & Query Helpers
# ---------------------------------------------------------------------------

T = t.TypeVar("T", bound=t.Any)
P = t.ParamSpec("P")
R = t.TypeVar("R")

T_MetaInstance = t.TypeVar("T_MetaInstance", bound="TypedTable")
T_Query = t.Union[
    "Table",
    Query,
    bool,
    None,
    "TypedTable",
    t.Type["TypedTable"],
    Expression,
]

T_subclass = t.TypeVar("T_subclass", "TypedTable", Table)
T_Field: t.TypeAlias = t.Union["TypedField[t.Any]", "Table", t.Type["TypedTable"]]

# use typing.cast(type, ...) to make mypy happy with unions
T_Value = t.TypeVar("T_Value")  # actual type of the Field (via Generic)

# table-ish parameter:
P_Table = t.Union[t.Type["TypedTable"], pydal.objects.Table]

Condition: t.TypeAlias = t.Optional[t.Callable[[P_Table, P_Table], Query | bool]]

OnQuery: t.TypeAlias = t.Optional[t.Callable[[P_Table, P_Table], list[Expression]]]

CacheModel = t.Callable[[str, CacheFn, int], Rows]
CacheTuple = tuple[CacheModel, int]

OrderBy: t.TypeAlias = str | Expression
GroupBy: t.TypeAlias = Field | Expression
Having: t.TypeAlias = Query | Expression

T_annotation = t.Type[t.Any] | types.UnionType



import tempfile

import pytest
from typer.testing import CliRunner

from src.typedal.__about__ import __version__
from src.typedal.cli import app, get_output_format

# by default, click's cli runner mixes stdout and stderr for some reason...
runner = CliRunner()


def test_version():
    result = runner.invoke(app, ["--version"])
    assert result.exit_code == 0
    assert __version__ in result.stdout


def test_generate_dry():
    result = runner.invoke(app, ["migrations.generate", "--dry-run"])
    assert result.exit_code == 0

    assert "would run" in result.stderr.lower()
    assert "create" in result.stderr

    with tempfile.NamedTemporaryFile() as f:
        f.write(b"...")

        result = runner.invoke(app, ["migrations.generate", "--dry-run", "--output-file", f.name])
        assert result.exit_code == 0
        assert "would run" in result.stderr.lower()
        assert "alter" in result.stderr


def test_stub():
    # mostly tested in pydal2sql already
    result = runner.invoke(app, ["migrations.stub", "my_test_migration", "--dry-run"])
    assert result.exit_code == 0

    assert "@migration" in result.stdout
    assert "def my_test_migration" in result.stdout


def test_run_dry():
    result = runner.invoke(app, ["migrations.run", "--dry-run"])
    assert result.exit_code == 0

    print(result.stdout)
    print(result.stderr)

    assert "would run" in result.stderr.lower()


def test_show_config():
    result = runner.invoke(app, ["--show-config"])
    assert result.exit_code == 0
    assert not result.stderr
    assert result.stdout.strip().startswith("<TypeDAL")


def test_get_output_format(capsys):
    import json

    import tomli
    import yaml

    with pytest.raises(ValueError):
        assert not get_output_format("bleepbloop")

    get_output_format("json")({"some": {"nested": "data"}})
    captured = capsys.readouterr()
    result = json.loads(captured.out)
    assert result
    assert result["some"]["nested"] == "data"

    get_output_format("yaml")({"some": {"nested": "data"}})
    captured = capsys.readouterr()
    result = yaml.load(captured.out, yaml.Loader)
    assert result
    assert result["some"]["nested"] == "data"

    get_output_format("toml")({"some": {"nested": "data"}})
    captured = capsys.readouterr()
    result = tomli.loads(captured.out)
    assert result
    assert result["some"]["nested"] == "data"

    plaintext = get_output_format("plaintext")
    assert plaintext
    plaintext({"some": {"nested": "data"}})
    captured = capsys.readouterr()
    assert captured.out
    assert not captured.err



import json
import tempfile

from contextlib_chdir import chdir
from pydal.validators import IS_EMAIL, IS_NOT_IN_DB

from src.typedal import TypedTable
from src.typedal.for_py4web import DAL, AuthUser, setup_py4web_tables
from src.typedal.serializers import as_json
from typedal.config import TypeDALConfig

db = DAL("sqlite:memory")


def test_P4W():
    assert hasattr(DAL, "on_request")
    assert hasattr(db, "on_request")


def test_serialize():
    @db.define()
    class JsonTable(TypedTable):
        field: str

    row = JsonTable.insert(field="Hey")

    l = json.loads(as_json.encode(JsonTable))
    assert l
    assert isinstance(l, dict)
    l = json.loads(as_json.encode(row))
    assert isinstance(l, dict)
    assert l

    l = json.loads(as_json.encode(JsonTable.paginate(limit=1)))

    assert l["pagination"]
    assert l["data"]


def test_auth_user():
    setup_py4web_tables(db)
    requirements = AuthUser.email.requires

    assert requirements
    assert isinstance(requirements, list)
    assert len(requirements) == 2

    assert isinstance(requirements[0], IS_EMAIL)
    assert isinstance(requirements[1], IS_NOT_IN_DB)


def test_py4web_dal_singleton():
    with tempfile.TemporaryDirectory() as d:
        with chdir(d):
            # note: typedal caching is disabled here because otherwise the cache table might send logs to {d}/log.sql
            # which can lead to problems in other tests :(
            db_1a = DAL("sqlite:memory", enable_typedal_caching=False)
            db_1b = DAL("sqlite:memory", enable_typedal_caching=False)

            db_2a = DAL("sqlite://test_py4web_dal_singleton", folder=d, enable_typedal_caching=False)
            db_2b = DAL("sqlite://test_py4web_dal_singleton", folder=d, enable_typedal_caching=False)

            conf = {"database": "sqlite:memory", "dialect": "sqlite", "pyproject": "", "flag_location": f"{d}/flags"}
            db_3a = DAL(config=TypeDALConfig.load(conf))
            db_3b = DAL(config=TypeDALConfig.load(conf))

            assert db_1a is db_1b
            assert db_1a._uri == db_1b._uri
            assert db_1a._db_uid == db_1b._db_uid

            assert db_2a is db_2b
            assert db_2a._uri == db_2b._uri
            assert db_2a._db_uid == db_2b._db_uid

            assert db_1a is not db_2a
            assert db_1a._uri != db_2a._uri
            assert db_1a._db_uid != db_2a._db_uid

            assert db_1b is not db_2b
            assert db_1b._uri != db_2b._uri
            assert db_1b._db_uid != db_2b._db_uid

            assert db_3a is db_3b
            assert db_3a is not db_1a
            assert db_3a is not db_2a

            assert db_1a._uri == db_1b._uri
            assert db_3a._uri == db_1a._uri
            assert db_3a._uri != db_2a._uri

    # reset singletons for later use:
    DAL._clear()



"""
Helpers to facilitate db-based caching.
"""

import contextlib
import datetime as dt
import hashlib
import json
import typing as t

import dill  # nosec
from pydal.objects import Field, Rows, Set

from .fields import TypedField
from .helpers import throw
from .rows import TypedRows
from .tables import TypedTable
from .types import CacheStatus, Query

if t.TYPE_CHECKING:
    from .core import TypeDAL
    from .query_builder import QueryBuilder


def get_now(tz: dt.timezone = dt.timezone.utc) -> dt.datetime:
    """
    Get the default datetime, optionally in a specific timezone.
    """
    return dt.datetime.now(tz)


class _TypedalCache(TypedTable):
    """
    Internal table to store previously loaded models.
    """

    key: TypedField[str]
    data: TypedField[bytes]
    cached_at = TypedField(dt.datetime, default=get_now)
    expires_at: TypedField[dt.datetime | None]


class _TypedalCacheDependency(TypedTable):
    """
    Internal table that stores dependencies to invalidate cache when a related table is updated.
    """

    entry: TypedField[_TypedalCache]
    table: TypedField[str]
    idx: TypedField[int]


def prepare(field: t.Any) -> str:
    """
    Prepare data to be used in a cache key.

    By sorting and stringifying data, queries can be syntactically different from each other \
        but when semantically exactly the same will still be loaded from cache.
    """
    if isinstance(field, str):
        return field
    elif isinstance(field, (dict, t.Mapping)):
        data = {str(k): prepare(v) for k, v in field.items()}
        return json.dumps(data, sort_keys=True)
    elif isinstance(field, t.Iterable):
        return ",".join(sorted([prepare(_) for _ in field]))
    elif isinstance(field, bool):
        return str(int(field))
    else:
        return str(field)


def create_cache_key(*fields: t.Any) -> str:
    """
    Turn any fields of data into a string.
    """
    return "|".join(prepare(field) for field in fields)


def hash_cache_key(cache_key: str | bytes) -> str:
    """
    Hash the input cache key with SHA 256.
    """
    h = hashlib.sha256()
    h.update(cache_key.encode() if isinstance(cache_key, str) else cache_key)
    return h.hexdigest()


def create_and_hash_cache_key(*fields: t.Any) -> tuple[str, str]:
    """
    Combine the input fields into one key and hash it with SHA 256.
    """
    key = create_cache_key(*fields)
    return key, hash_cache_key(key)


DependencyTuple = tuple[str, int]  # table + id
DependencyTupleSet = set[DependencyTuple]


def _get_table_name(field: Field) -> str:
    """
    Get the table name from a field or alias.
    """
    return str(field._table).split(" AS ")[0].strip()


def _get_dependency_ids(rows: Rows, dependency_keys: list[tuple[Field, str]]) -> DependencyTupleSet:
    dependencies = set()
    for row in rows:
        for field, table in dependency_keys:
            if idx := row[field]:
                dependencies.add((table, idx))

    return dependencies


def _determine_dependencies_auto(rows: Rows) -> DependencyTupleSet:
    dependency_keys = []
    for field in rows.fields:
        if str(field).endswith(".id"):
            table_name = _get_table_name(field)

            dependency_keys.append((field, table_name))

    return _get_dependency_ids(rows, dependency_keys)


def _determine_dependencies(instance: TypedRows[t.Any], rows: Rows, depends_on: list[t.Any]) -> DependencyTupleSet:
    if not depends_on:
        return _determine_dependencies_auto(rows)

    target_field_names = set()
    for field in depends_on:
        if "." not in field:
            field = f"{instance.model._table}.{field}"

        target_field_names.add(str(field))

    dependency_keys = []
    for field in rows.fields:
        if str(field) in target_field_names:
            table_name = _get_table_name(field)

            dependency_keys.append((field, table_name))

    return _get_dependency_ids(rows, dependency_keys)


def remove_cache(idx: int | t.Iterable[int], table: str) -> None:
    """
    Remove any cache entries that are dependant on one or multiple indices of a table.
    """
    if not isinstance(idx, t.Iterable):
        idx = [idx]

    related = (
        _TypedalCacheDependency.where(table=table).where(lambda row: row.idx.belongs(idx)).select("entry").to_sql()
    )

    _TypedalCache.where(_TypedalCache.id.belongs(related)).delete()


def remove_cache_for_table(table: str) -> None:
    """
    Remove all cache entries that depend on a table.

    Used for inserts where we don't know which cached queries
    the new row would match.
    """
    related = _TypedalCacheDependency.where(table=table).select("entry").to_sql()
    _TypedalCache.where(_TypedalCache.id.belongs(related)).delete()


def clear_cache() -> None:
    """
    Remove everything from the cache.

    Immediately commits
    """
    db: TypeDAL = _TypedalCache._db or throw(
        RuntimeError("@define or db.define is not called on typedal caching classes yet!")
    )

    _TypedalCache.truncate("RESTART IDENTITY CASCADE")
    db.commit()


def clear_expired() -> int:
    """
    Remove all expired items from the cache.

    By default, expired items are only removed when trying to access them.
    """
    now = get_now()
    return len(_TypedalCache.where(_TypedalCache.expires_at != None).where(_TypedalCache.expires_at < now).delete())


def _remove_cache(s: Set, tablename: str) -> None:
    """
    Used as the table._before_update and table._after_update for every TypeDAL table (on by default).
    """
    indeces = s.select("id").column("id")
    remove_cache(indeces, tablename)


T_TypedTable = t.TypeVar("T_TypedTable", bound=TypedTable)


def get_expire(
    expires_at: t.Optional[dt.datetime] = None,
    ttl: t.Optional[int | dt.timedelta] = None,
    now: t.Optional[dt.datetime] = None,
) -> dt.datetime | None:
    """
    Based on an expires_at date or a ttl (in seconds or a time delta), determine the expire date.
    """
    now = now or get_now()

    if expires_at and ttl:
        raise ValueError("Please only supply an `expired at` date or a `ttl` in seconds!")
    elif isinstance(ttl, dt.timedelta):
        return now + ttl
    elif ttl:
        return now + dt.timedelta(seconds=ttl)
    elif expires_at:
        return expires_at

    return None


def _insert_cache_entry(
    db: "TypeDAL",
    key: str,
    data: t.Any,
    expires_at: dt.datetime | None,
    deps: DependencyTupleSet,
) -> None:
    """
    Shared internal function to insert cache entry and dependencies.
    """
    entry = _TypedalCache.insert(
        key=key,
        data=dill.dumps(data),
        expires_at=expires_at,
    )

    _TypedalCacheDependency.bulk_insert([{"entry": entry, "table": table, "idx": idx} for table, idx in deps])

    db.commit()


def save_to_cache(
    instance: TypedRows[T_TypedTable],
    rows: Rows,
    expires_at: t.Optional[dt.datetime] = None,
    ttl: t.Optional[int | dt.timedelta] = None,
) -> TypedRows[T_TypedTable]:
    """
    Save a typedrows result to the database, and save dependencies from rows.

    You can call .cache(...) with dependent fields (e.g. User.id) or this function will determine them automatically.
    """
    db = rows.db
    if (c := instance.metadata.get("cache", {})) and c.get("enabled") and (key := c.get("key")):
        expires_at = get_expire(expires_at=expires_at, ttl=ttl) or c.get("expires_at")
        deps = _determine_dependencies(instance, rows, c["depends_on"])

        _insert_cache_entry(db, key, instance, expires_at, deps)

        instance.metadata["cache"]["status"] = "fresh"
    return instance


class CacheMiss:
    """Sentinel class to represent a cache miss, distinguishing it from a None value."""

    def __bool__(self) -> bool:  # pragma: no cover
        """
        Ensures the sentinel evaluates to False in boolean contexts.

        Example:
            res = _load_memoize_from_cache("key")
            if not res:  # Triggers if a CacheMiss occurs
                ...
        """
        return False


_CACHE_SENTINEL: t.Final[CacheMiss] = CacheMiss()


def _fetch_cached_payload(key: str) -> tuple[t.Any, t.Any] | None:
    """
    Retrieves and validates a cache entry from the database.

    If the entry exists but is expired, it is deleted from the database
    to maintain storage hygiene.

    Args:
        key: The unique string identifier for the cache entry.

    Returns:
        A tuple of (deserialized_data, db_row) if valid; None if missing or expired.
    """
    row = _TypedalCache.where(key=key).first()
    if not row:
        return None

    now = get_now()
    # Ensure comparison is offset-aware if the row has a timestamp
    expires = row.expires_at.replace(tzinfo=dt.timezone.utc) if row.expires_at else None

    if expires and now >= expires:
        row.delete_record()
        return None

    # Only one place for deserialization to happen
    return dill.loads(row.data), row  # nosec


def _load_from_cache(key: str, db: "TypeDAL") -> t.Any | None:
    """
    Loads a specific TypeDAL model instance from the cache and re-hydrates it.

    This binds the cached object back to the active database connection and
    restores its instance methods/metadata.

    Args:
        key: Cache key.
        db: The active TypeDAL database instance.

    Returns:
        The re-hydrated model instance, or None if the load fails.
    """
    result = _fetch_cached_payload(key)
    if not result:
        return None

    inst, row = result

    # Re-hydrate the model instance with metadata and DB context
    inst.metadata.setdefault("cache", {}).update(
        {"status": "cached", "cached_at": row.cached_at, "expires_at": row.expires_at},
    )

    inst.db = db
    inst.model = db._class_map[inst.model]
    inst.model._setup_instance_methods(inst.model)
    return inst


def _load_memoize_from_cache(key: str) -> t.Any:
    """
    Low-level retrieval for memoized results.

    Used when the caller doesn't need TypeDAL model re-hydration, just the raw data.

    Args:
        key: Cache key.

    Returns:
        The deserialized data or the _CACHE_SENTINEL object if not found.
    """
    with contextlib.suppress(Exception):
        if result := _fetch_cached_payload(key):
            return result[0]

    return _CACHE_SENTINEL


def load_from_cache(key: str, db: "TypeDAL") -> t.Any | None:
    """
    Public entry point to load model instances from cache.

    Wraps the internal loader in a broad exception handler to ensure that
    cache failures never crash the main application flow.

    Args:
        key: Cache key.
        db: The active TypeDAL database instance.

    Returns:
        The model instance or None on any failure/miss.
    """
    try:
        return _load_from_cache(key, db)
    except Exception:  # pragma: no cover
        return None


def humanize_bytes(size: int | float) -> str:
    """
    Turn a number of bytes into a human-readable version (e.g. 124 GB).
    """
    if not size:
        return "0"

    suffixes = ["B", "KB", "MB", "GB", "TB", "PB"]  # List of suffixes for different magnitudes
    suffix_index = 0

    while size > 1024 and suffix_index < len(suffixes) - 1:
        suffix_index += 1
        size /= 1024.0

    return f"{size:.2f} {suffixes[suffix_index]}"


def _expired_and_valid_query() -> tuple[str, str]:
    expired_items = (
        _TypedalCache.where(lambda row: (row.expires_at < get_now()) & (row.expires_at != None))
        .select(_TypedalCache.id)
        .to_sql()
    )

    valid_items = _TypedalCache.where(~_TypedalCache.id.belongs(expired_items)).select(_TypedalCache.id).to_sql()

    return expired_items, valid_items


T = t.TypeVar("T")
Stats = t.TypedDict("Stats", {"total": T, "valid": T, "expired": T})

RowStats = t.TypedDict(
    "RowStats",
    {
        "Dependent Cache Entries": int,
    },
)


def _row_stats(db: "TypeDAL", table: str, query: Query) -> RowStats:
    count_field = _TypedalCacheDependency.entry.count()
    stats: TypedRows[_TypedalCacheDependency] = db(query & (_TypedalCacheDependency.table == table)).select(
        _TypedalCacheDependency.entry,
        count_field,
        groupby=_TypedalCacheDependency.entry,
    )
    return {
        "Dependent Cache Entries": len(stats),
    }


def row_stats(db: "TypeDAL", table: str, row_id: str) -> Stats[RowStats]:
    """
    Collect caching stats for a specific table row (by ID).
    """
    expired_items, valid_items = _expired_and_valid_query()

    query = _TypedalCacheDependency.idx == row_id

    return {
        "total": _row_stats(db, table, query),
        "valid": _row_stats(db, table, _TypedalCacheDependency.entry.belongs(valid_items) & query),
        "expired": _row_stats(db, table, _TypedalCacheDependency.entry.belongs(expired_items) & query),
    }


TableStats = t.TypedDict(
    "TableStats",
    {
        "Dependent Cache Entries": int,
        "Associated Table IDs": int,
    },
)


def _table_stats(db: "TypeDAL", table: str, query: Query) -> TableStats:
    count_field = _TypedalCacheDependency.entry.count()
    stats: TypedRows[_TypedalCacheDependency] = db(query & (_TypedalCacheDependency.table == table)).select(
        _TypedalCacheDependency.entry,
        count_field,
        groupby=_TypedalCacheDependency.entry,
    )
    return {
        "Dependent Cache Entries": len(stats),
        "Associated Table IDs": sum(stats.column(count_field)),
    }


def table_stats(db: "TypeDAL", table: str) -> Stats[TableStats]:
    """
    Collect caching stats for a table.
    """
    expired_items, valid_items = _expired_and_valid_query()

    return {
        "total": _table_stats(db, table, _TypedalCacheDependency.id > 0),
        "valid": _table_stats(db, table, _TypedalCacheDependency.entry.belongs(valid_items)),
        "expired": _table_stats(db, table, _TypedalCacheDependency.entry.belongs(expired_items)),
    }


GenericStats = t.TypedDict(
    "GenericStats",
    {
        "entries": int,
        "dependencies": int,
        "size": str,
    },
)


def _calculate_stats(db: "TypeDAL", query: Query) -> GenericStats:
    sum_len_field = _TypedalCache.data.len().sum()
    size_row = db(query).select(sum_len_field).first()

    size = size_row[sum_len_field] if size_row else 0

    return {
        "entries": _TypedalCache.where(query).count(),
        "dependencies": db(_TypedalCacheDependency.entry.belongs(query)).count(),
        "size": humanize_bytes(size),
    }


def calculate_stats(db: "TypeDAL") -> Stats[GenericStats]:
    """
    Collect generic caching stats.
    """
    expired_items, valid_items = _expired_and_valid_query()

    return {
        "total": _calculate_stats(db, _TypedalCache.id > 0),
        "valid": _calculate_stats(db, _TypedalCache.id.belongs(valid_items)),
        "expired": _calculate_stats(db, _TypedalCache.id.belongs(expired_items)),
    }


def memoize(
    db: "TypeDAL",
    func: t.Callable[..., T],
    *args: TypedRows[t.Any] | TypedTable,
    key: str | None = None,
    ttl: int | dt.timedelta | dt.datetime | None = None,
    **kwargs: t.Any,
) -> tuple[T, CacheStatus]:
    """
    Cache the result of a function applied to TypedRow(s).

    Tracks dependencies on the table(s) so the cache invalidates
    when those rows are updated/deleted.

    Args:
        db: TypeDAL database
        func: Function to cache
        *args: TypedRow/TypedRows instances only
        key: Cache key (required for lambdas)
        ttl: Time to live in seconds/timedelta, or datetime to expire at
        **kwargs: Extra parameters passed to func

    Returns:
        tuple of (result, cache_status)
    """
    if not key:
        if func.__name__ == "<lambda>":
            raise ValueError(
                "Lambda functions require explicit 'key' parameter. Use: db.memoize(your_func, data, key='my_key')",
            )
        key = func.__qualname__

    # Extract dependencies from args
    deps: DependencyTupleSet = set()
    for arg in args:
        if isinstance(arg, TypedRows):
            for row in arg:
                deps.add((str(row._table), row.id))
        elif isinstance(arg, TypedTable):
            deps.add((str(arg._table), arg.id))

    # Generate cache key
    _, hashed_key = create_and_hash_cache_key(key, *[getattr(arg, "id", None) for arg in args], kwargs)

    # Try to load from cache
    cached = _load_memoize_from_cache(hashed_key)
    if cached is not _CACHE_SENTINEL:
        return cached, "cached"
    # Cache miss - compute result

    def track_execute(qb: "QueryBuilder[t.Any]", raw: Rows):
        # find dependant table+id combinations, includes relationships:
        deps.update(_determine_dependencies_auto(raw))

        # tables: qb.model;
        # something with qb.relationships
        # something with qb.select_args

        related_tables = (
            {
                # original table
                str(qb.model)
            }
            | {
                # other tables in select()
                _get_table_name(arg)
                for arg in qb.select_args
            }
            | {
                # other tables in relationships
                str(r.table)
                for r in qb.relationships.values()
            }
        )

        # mark dependency for every relevant table in this query without id:
        deps.update({(table, 0) for table in related_tables})

    def track_collect(qb: "QueryBuilder[t.Any]", _: TypedRows[t.Any], raw: Rows) -> None:
        return track_execute(qb, raw)

    # hooks every .collect() to track extra dependencies
    db._after_collect.append(track_collect)
    db._after_execute.append(track_execute)
    try:
        result = func(*args, **kwargs)
    finally:
        db._after_collect.remove(track_collect)
        db._after_execute.remove(track_execute)

    # Save to cache
    if isinstance(ttl, dt.datetime):
        expires_at: dt.datetime | None = ttl
    else:
        expires_at = get_expire(ttl=ttl)

    _insert_cache_entry(db, hashed_key, result, expires_at, deps)

    return result, "fresh"



"""
Contains base functionality related to Tables.
"""

from __future__ import annotations

import copy
import csv
import functools
import json
import typing as t
import uuid

import pydal.objects
from pydal._globals import DEFAULT

from .constants import JOIN_OPTIONS
from .core import TypeDAL
from .helpers import classproperty, throw
from .serializers import as_json
from .types import (
    AnyDict,
    Condition,
    Expression,
    Field,
    OnQuery,
    OpRow,
    OrderBy,
    P,
    Query,
    R,
    Reference,
    Row,
    SelectKwargs,
    Set,
    T,
    T_MetaInstance,
    T_Query,
    Table,
)

if t.TYPE_CHECKING:
    from .relationships import Relationship
    from .rows import PaginatedRows, TypedRows


def reorder_fields(
    table: pydal.objects.Table,
    fields: t.Iterable[str | TypedField[t.Any] | Field],
    keep_others: bool = True,
) -> None:
    """
    Reorder fields of a pydal table.

    Args:
        table: The pydal table object (e.g., db.mytable).
        fields: List of field names (str) or Field objects in desired order.
        keep_others (bool):
            - True (default): keep other fields at the end, in their original order.
            - False: remove other fields (only keep what's specified).
    """
    # Normalize input to field names
    desired = [f.name if isinstance(f, (TypedField, Field, pydal.objects.Field)) else str(f) for f in fields]

    new_order = [f for f in desired if f in table._fields]

    if keep_others:
        # Start with desired fields, then append the rest
        new_order.extend(f for f in table._fields if f not in desired)

    table._fields = new_order


class TableMeta(type):
    """
    This metaclass contains functionality on table classes, that doesn't exist on its instances.

    Example:
        class MyTable(TypedTable):
            some_field: TypedField[int]

    MyTable.update_or_insert(...) # should work

    MyTable.some_field # -> Field, can be used to query etc.

    row = MyTable.first() # returns instance of MyTable

    # row.update_or_insert(...) # shouldn't work!

    row.some_field # -> int, with actual data

    """

    # set up by db.define:
    # _db: TypeDAL | None = None
    # _table: Table | None = None
    _db: TypeDAL | None = None
    _table: Table | None = None
    _relationships: dict[str, Relationship[t.Any]] | None = None

    #########################
    # TypeDAL custom logic: #
    #########################

    def __set_internals__(self, db: pydal.DAL, table: Table, relationships: dict[str, Relationship[t.Any]]) -> None:
        """
        Store the related database and pydal table for later usage.
        """
        self._db = db
        self._table = table
        self._relationships = relationships

    def __getattr__(self, col: str) -> t.Optional[Field]:
        """
        Magic method used by TypedTableMeta to get a database field with dot notation on a class.

        Example:
            SomeTypedTable.col -> db.table.col (via TypedTableMeta.__getattr__)

        """
        if self._table:
            return getattr(self._table, col, None)

        return None

    def _ensure_table_defined(self) -> Table:
        if not self._table:
            raise EnvironmentError("@define or db.define is not called on this class yet!")
        return self._table

    def __iter__(self) -> t.Generator[Field, None, None]:
        """
        Loop through the columns of this model.
        """
        table = self._ensure_table_defined()
        yield from iter(table)

    def __getitem__(self, item: str) -> Field:
        """
        Allow dict notation to get a column of this table (-> Field instance).
        """
        table = self._ensure_table_defined()
        return table[item]

    def __str__(self) -> str:
        """
        Normally, just returns the underlying table name, but with a fallback if the model is unbound.
        """
        if self._table:
            return str(self._table)
        else:
            return f"<unbound table {self.__name__}>"

    def from_row(self: t.Type[T_MetaInstance], row: pydal.objects.Row) -> T_MetaInstance:
        """
        Create a model instance from a pydal row.
        """
        return self(row)

    def all(self: t.Type[T_MetaInstance]) -> "TypedRows[T_MetaInstance]":
        """
        Return all rows for this model.
        """
        return self.collect()

    def get_relationships(self) -> dict[str, Relationship[t.Any]]:
        """
        Return the registered relationships of the current model.
        """
        return self._relationships or {}

    ##########################
    # TypeDAL Modified Logic #
    ##########################

    def insert(self: t.Type[T_MetaInstance], **fields: t.Any) -> T_MetaInstance:
        """
        This is only called when db.define is not used as a decorator.

        cls.__table functions as 'self'

        Args:
            **fields: t.Anything you want to insert in the database

        Returns: the ID of the new row.

        """
        table = self._ensure_table_defined()

        result = table.insert(**fields)
        # it already is an int but mypy doesn't understand that
        return self(result)

    def _insert(self, **fields: t.Any) -> str:
        table = self._ensure_table_defined()

        return str(table._insert(**fields))

    def bulk_insert(self: t.Type[T_MetaInstance], items: list[AnyDict]) -> "TypedRows[T_MetaInstance]":
        """
        Insert multiple rows, returns a TypedRows set of new instances.
        """
        table = self._ensure_table_defined()
        result = table.bulk_insert(items)
        return self.where(lambda row: row.id.belongs(result)).collect()

    def update_or_insert(
        self: t.Type[T_MetaInstance],
        query: T_Query | AnyDict = DEFAULT,
        **values: t.Any,
    ) -> T_MetaInstance:
        """
        Update a row if query matches, else insert a new one.

        Returns the created or updated instance.
        """
        table = self._ensure_table_defined()

        if query is DEFAULT:
            record = table(**values)
        elif isinstance(query, dict):
            record = table(**query)
        else:
            record = table(query)

        if not record:
            return self.insert(**values)

        record.update_record(**values)
        return self(record)

    def validate_and_insert(
        self: t.Type[T_MetaInstance],
        **fields: t.Any,
    ) -> tuple[t.Optional[T_MetaInstance], t.Optional[dict[str, str]]]:
        """
        Validate input data and then insert a row.

        Returns a tuple of (the created instance, a dict of errors).
        """
        table = self._ensure_table_defined()
        result = table.validate_and_insert(**fields)
        if row_id := result.get("id"):
            return self(row_id), None
        else:
            return None, result.get("errors")

    def validate_and_update(
        self: t.Type[T_MetaInstance],
        query: Query,
        **fields: t.Any,
    ) -> tuple[t.Optional[T_MetaInstance], t.Optional[dict[str, str]]]:
        """
        Validate input data and then update max 1 row.

        Returns a tuple of (the updated instance, a dict of errors).
        """
        table = self._ensure_table_defined()

        result = table.validate_and_update(query, **fields)

        if errors := result.get("errors"):
            return None, errors
        elif row_id := result.get("id"):
            return self(row_id), None
        else:  # pragma: no cover
            # update on query without result (shouldnt happen)
            return None, None

    def validate_and_update_or_insert(
        self: t.Type[T_MetaInstance],
        query: Query,
        **fields: t.Any,
    ) -> tuple[t.Optional[T_MetaInstance], t.Optional[dict[str, str]]]:
        """
        Validate input data and then update_and_insert (on max 1 row).

        Returns a tuple of (the updated/created instance, a dict of errors).
        """
        table = self._ensure_table_defined()
        result = table.validate_and_update_or_insert(query, **fields)

        if errors := result.get("errors"):
            return None, errors
        elif row_id := result.get("id"):
            return self(row_id), None
        else:  # pragma: no cover
            # update on query without result (shouldnt happen)
            return None, None

    def select(self: t.Type[T_MetaInstance], *a: t.Any, **kw: t.Any) -> "QueryBuilder[T_MetaInstance]":
        """
        See QueryBuilder.select!
        """
        return QueryBuilder(self).select(*a, **kw)

    def column(self: t.Type[T_MetaInstance], field: T | TypedField[T], **options: t.Unpack[SelectKwargs]) -> list[T]:
        """
        Get all values in a specific column.

        Shortcut for `.select(field).execute().column(field)`.
        """
        return QueryBuilder(self).select(field, **options).execute().column(field)

    def paginate(self: t.Type[T_MetaInstance], limit: int, page: int = 1) -> "PaginatedRows[T_MetaInstance]":
        """
        See QueryBuilder.paginate!
        """
        return QueryBuilder(self).paginate(limit=limit, page=page)

    def chunk(self: t.Type[T_MetaInstance], chunk_size: int) -> t.Generator["TypedRows[T_MetaInstance]", t.Any, None]:
        """
        See QueryBuilder.chunk!
        """
        return QueryBuilder(self).chunk(chunk_size)

    def where(self: t.Type[T_MetaInstance], *a: t.Any, **kw: t.Any) -> "QueryBuilder[T_MetaInstance]":
        """
        See QueryBuilder.where!
        """
        return QueryBuilder(self).where(*a, **kw)

    def orderby(self: t.Type[T_MetaInstance], *fields: OrderBy) -> "QueryBuilder[T_MetaInstance]":
        """
        See QueryBuilder.orderby!
        """
        return QueryBuilder(self).orderby(*fields)

    def groupby(self: t.Type[T_MetaInstance], *fields: t.Any) -> "QueryBuilder[T_MetaInstance]":
        """
        See QueryBuilder.groupby!
        """
        return QueryBuilder(self).groupby(*fields)

    def having(self: t.Type[T_MetaInstance], condition: t.Any) -> "QueryBuilder[T_MetaInstance]":
        """
        See QueryBuilder.having!
        """
        return QueryBuilder(self).having(condition)

    def cache(self: t.Type[T_MetaInstance], *deps: t.Any, **kwargs: t.Any) -> "QueryBuilder[T_MetaInstance]":
        """
        See QueryBuilder.cache!
        """
        return QueryBuilder(self).cache(*deps, **kwargs)

    def count(self: t.Type[T_MetaInstance]) -> int:
        """
        See QueryBuilder.count!
        """
        return QueryBuilder(self).count()

    def exists(self: t.Type[T_MetaInstance]) -> bool:
        """
        See QueryBuilder.exists!
        """
        return QueryBuilder(self).exists()

    def first(self: t.Type[T_MetaInstance]) -> T_MetaInstance | None:
        """
        See QueryBuilder.first!
        """
        return QueryBuilder(self).first()

    def first_or_fail(self: t.Type[T_MetaInstance]) -> T_MetaInstance:
        """
        See QueryBuilder.first_or_fail!
        """
        return QueryBuilder(self).first_or_fail()

    def join(
        self: t.Type[T_MetaInstance],
        *fields: str | t.Type[TypedTable] | Relationship[t.Any],
        method: JOIN_OPTIONS = None,
        on: OnQuery | list[Expression] | Expression = None,
        condition: Condition = None,
        condition_and: Condition = None,
    ) -> "QueryBuilder[T_MetaInstance]":
        """
        See QueryBuilder.join!
        """
        return QueryBuilder(self).join(*fields, on=on, condition=condition, method=method, condition_and=condition_and)

    def collect(self: t.Type[T_MetaInstance], verbose: bool = False) -> "TypedRows[T_MetaInstance]":
        """
        See QueryBuilder.collect!
        """
        return QueryBuilder(self).collect(verbose=verbose)

    @property
    def ALL(cls) -> pydal.objects.SQLALL:
        """
        Select all fields for this table.
        """
        table = cls._ensure_table_defined()

        return table.ALL

    ##########################
    # TypeDAL Shadowed Logic #
    ##########################
    fields: list[str]

    # other table methods:

    def truncate(self, mode: str = "") -> None:
        """
        Remove all data and reset index.
        """
        table = self._ensure_table_defined()
        table.truncate(mode)

    def drop(self, mode: str = "") -> None:
        """
        Remove the underlying table.
        """
        table = self._ensure_table_defined()
        table.drop(mode)

    def create_index(self, name: str, *fields: str | Field, **kwargs: t.Any) -> bool:
        """
        Add an index on some columns of this table.
        """
        table = self._ensure_table_defined()
        result = table.create_index(name, *fields, **kwargs)
        return t.cast(bool, result)

    def drop_index(self, name: str, if_exists: bool = False) -> bool:
        """
        Remove an index from this table.
        """
        table = self._ensure_table_defined()
        result = table.drop_index(name, if_exists)
        return t.cast(bool, result)

    def import_from_csv_file(
        self,
        csvfile: t.TextIO,
        id_map: dict[str, str] = None,
        null: t.Any = "<NULL>",
        unique: str = "uuid",
        id_offset: dict[str, int] = None,  # id_offset used only when id_map is None
        transform: t.Callable[[dict[t.Any, t.Any]], dict[t.Any, t.Any]] = None,
        validate: bool = False,
        encoding: str = "utf-8",
        delimiter: str = ",",
        quotechar: str = '"',
        quoting: int = csv.QUOTE_MINIMAL,
        restore: bool = False,
        **kwargs: t.Any,
    ) -> None:
        """
        Load a csv file into the database.
        """
        table = self._ensure_table_defined()
        table.import_from_csv_file(
            csvfile,
            id_map=id_map,
            null=null,
            unique=unique,
            id_offset=id_offset,
            transform=transform,
            validate=validate,
            encoding=encoding,
            delimiter=delimiter,
            quotechar=quotechar,
            quoting=quoting,
            restore=restore,
            **kwargs,
        )

    def on(self, query: bool | Query) -> Expression:
        """
        Shadow Table.on.

        Used for joins.

        See Also:
            http://web2py.com/books/default/chapter/29/06/the-database-abstraction-layer?search=export_to_csv_file#One-to-mt.Any-relation
        """
        table = self._ensure_table_defined()
        return t.cast(Expression, table.on(query))

    def with_alias(self: t.Type[T_MetaInstance], alias: str) -> t.Type[T_MetaInstance]:
        """
        Shadow Table.with_alias.

        Useful for joins when joining the same table multiple times.

        See Also:
            http://web2py.com/books/default/chapter/29/06/the-database-abstraction-layer#One-to-mt.Any-relation
        """
        table = self._ensure_table_defined()
        return t.cast(t.Type[T_MetaInstance], table.with_alias(alias))

    def unique_alias(self: t.Type[T_MetaInstance]) -> t.Type[T_MetaInstance]:
        """
        Generates a unique alias for this table.

        Useful for joins when joining the same table multiple times
            and you don't want to keep track of aliases yourself.
        """
        key = f"{self.__name__.lower()}_{hash(uuid.uuid4())}"
        return self.with_alias(key)

    # hooks:
    def _hook_once(
        cls: t.Type[T_MetaInstance],
        hooks: list[t.Callable[P, R]],
        fn: t.Callable[P, R],
    ) -> t.Type[T_MetaInstance]:
        @functools.wraps(fn)
        def wraps(*a: P.args, **kw: P.kwargs) -> R:
            try:
                return fn(*a, **kw)
            finally:
                hooks.remove(wraps)

        hooks.append(wraps)
        return cls

    def before_insert(
        cls: t.Type[T_MetaInstance],
        fn: t.Callable[[T_MetaInstance], t.Optional[bool]] | t.Callable[[OpRow], t.Optional[bool]],
    ) -> t.Type[T_MetaInstance]:
        """
        Add a before insert hook.
        """
        if fn not in cls._before_insert:
            cls._before_insert.append(fn)
        return cls

    def before_insert_once(
        cls: t.Type[T_MetaInstance],
        fn: t.Callable[[T_MetaInstance], t.Optional[bool]] | t.Callable[[OpRow], t.Optional[bool]],
    ) -> t.Type[T_MetaInstance]:
        """
        Add a before insert hook that only fires once and then removes itself.
        """
        return cls._hook_once(cls._before_insert, fn)  # type: ignore

    def after_insert(
        cls: t.Type[T_MetaInstance],
        fn: (
            t.Callable[[T_MetaInstance, Reference], t.Optional[bool]] | t.Callable[[OpRow, Reference], t.Optional[bool]]
        ),
    ) -> t.Type[T_MetaInstance]:
        """
        Add an after insert hook.
        """
        if fn not in cls._after_insert:
            cls._after_insert.append(fn)
        return cls

    def after_insert_once(
        cls: t.Type[T_MetaInstance],
        fn: (
            t.Callable[[T_MetaInstance, Reference], t.Optional[bool]] | t.Callable[[OpRow, Reference], t.Optional[bool]]
        ),
    ) -> t.Type[T_MetaInstance]:
        """
        Add an after insert hook that only fires once and then removes itself.
        """
        return cls._hook_once(cls._after_insert, fn)  # type: ignore

    def before_update(
        cls: t.Type[T_MetaInstance],
        fn: t.Callable[[Set, T_MetaInstance], t.Optional[bool]] | t.Callable[[Set, OpRow], t.Optional[bool]],
    ) -> t.Type[T_MetaInstance]:
        """
        Add a before update hook.
        """
        if fn not in cls._before_update:
            cls._before_update.append(fn)
        return cls

    def before_update_once(
        cls,
        fn: t.Callable[[Set, T_MetaInstance], t.Optional[bool]] | t.Callable[[Set, OpRow], t.Optional[bool]],
    ) -> t.Type[T_MetaInstance]:
        """
        Add a before update hook that only fires once and then removes itself.
        """
        return cls._hook_once(cls._before_update, fn)  # type: ignore

    def after_update(
        cls: t.Type[T_MetaInstance],
        fn: t.Callable[[Set, T_MetaInstance], t.Optional[bool]] | t.Callable[[Set, OpRow], t.Optional[bool]],
    ) -> t.Type[T_MetaInstance]:
        """
        Add an after update hook.
        """
        if fn not in cls._after_update:
            cls._after_update.append(fn)
        return cls

    def after_update_once(
        cls: t.Type[T_MetaInstance],
        fn: t.Callable[[Set, T_MetaInstance], t.Optional[bool]] | t.Callable[[Set, OpRow], t.Optional[bool]],
    ) -> t.Type[T_MetaInstance]:
        """
        Add an after update hook that only fires once and then removes itself.
        """
        return cls._hook_once(cls._after_update, fn)  # type: ignore

    def before_delete(cls: t.Type[T_MetaInstance], fn: t.Callable[[Set], t.Optional[bool]]) -> t.Type[T_MetaInstance]:
        """
        Add a before delete hook.
        """
        if fn not in cls._before_delete:
            cls._before_delete.append(fn)
        return cls

    def before_delete_once(
        cls: t.Type[T_MetaInstance],
        fn: t.Callable[[Set], t.Optional[bool]],
    ) -> t.Type[T_MetaInstance]:
        """
        Add a before delete hook that only fires once and then removes itself.
        """
        return cls._hook_once(cls._before_delete, fn)

    def after_delete(cls: t.Type[T_MetaInstance], fn: t.Callable[[Set], t.Optional[bool]]) -> t.Type[T_MetaInstance]:
        """
        Add an after delete hook.
        """
        if fn not in cls._after_delete:
            cls._after_delete.append(fn)
        return cls

    def after_delete_once(
        cls: t.Type[T_MetaInstance],
        fn: t.Callable[[Set], t.Optional[bool]],
    ) -> t.Type[T_MetaInstance]:
        """
        Add an after delete hook that only fires once and then removes itself.
        """
        return cls._hook_once(cls._after_delete, fn)

    def reorder_fields(cls, *fields: str | Field | TypedField[t.Any], keep_others: bool = True) -> None:
        """
        Reorder fields of a typedal table.

        Args:
            fields: List of field names (str) or Field objects in desired order.
            keep_others (bool):
                - True (default): keep other fields at the end, in their original order.
                - False: remove other fields (only keep what's specified).
        """
        return reorder_fields(cls._table, fields, keep_others=keep_others)


class _TypedTable:
    """
    This class is a final shared parent between TypedTable and Mixins.

    This needs to exist because otherwise the __on_define__ of Mixins are not executed.
    Notably, this class exists at a level ABOVE the `metaclass=TableMeta`,
        because otherwise typing gets confused when Mixins are used and multiple types could satisfy
            generic 'T subclass of TypedTable'
        -> Setting 'TypedTable' as the parent for Mixin does not work at runtime (and works semi at type check time)
    """

    id: "TypedField[int]"

    _before_insert: list[t.Callable[[t.Self], t.Optional[bool]] | t.Callable[[OpRow], t.Optional[bool]]]
    _after_insert: list[
        t.Callable[[t.Self, Reference], t.Optional[bool]] | t.Callable[[OpRow, Reference], t.Optional[bool]]
    ]
    _before_update: list[t.Callable[[Set, t.Self], t.Optional[bool]] | t.Callable[[Set, OpRow], t.Optional[bool]]]
    _after_update: list[t.Callable[[Set, t.Self], t.Optional[bool]] | t.Callable[[Set, OpRow], t.Optional[bool]]]
    _before_delete: list[t.Callable[[Set], t.Optional[bool]]]
    _after_delete: list[t.Callable[[Set], t.Optional[bool]]]

    @classmethod
    def __on_define__(cls, db: TypeDAL) -> None:
        """
        Method that can be implemented by tables to do an action after db.define is completed.

        This can be useful if you need to add something like requires=IS_NOT_IN_DB(db, "table.field"),
        where you need a reference to the current database, which may not exist yet when defining the model.
        """

    @classproperty
    def _hooks(cls) -> dict[str, list[t.Callable[..., t.Optional[bool]]]]:
        return {
            "before_insert": cls._before_insert,
            "after_insert": cls._after_insert,
            "before_update": cls._before_update,
            "after_update": cls._after_update,
            "before_delete": cls._before_delete,
            "after_delete": cls._after_delete,
        }


class TypedTable(_TypedTable, metaclass=TableMeta):
    """
    Enhanded modeling system on top of pydal's Table that adds typing and additional functionality.
    """

    # set up by 'new':
    _row: Row | None = None
    _rows: tuple[Row, ...] = ()

    _with: list[str]

    def _setup_instance_methods(self) -> None:
        self.as_dict = self._as_dict  # type: ignore
        self.__json__ = self.as_json = self._as_json  # type: ignore
        # self.as_yaml = self._as_yaml  # type: ignore
        self.as_xml = self._as_xml  # type: ignore

        self.update = self._update  # type: ignore

        self.delete_record = self._delete_record  # type: ignore
        self.update_record = self._update_record  # type: ignore

    def __new__(
        cls,
        row_or_id: t.Union[Row, Query, pydal.objects.Set, int, str, None, "TypedTable"] = None,
        **filters: t.Any,
    ) -> t.Self:
        """
        Create a Typed Rows model instance from an existing row, ID or query.

        Examples:
            MyTable(1)
            MyTable(id=1)
            MyTable(MyTable.id == 1)
        """
        table = cls._ensure_table_defined()
        inst = super().__new__(cls)

        if isinstance(row_or_id, TypedTable):
            # existing typed table instance!
            return t.cast(t.Self, row_or_id)

        elif isinstance(row_or_id, pydal.objects.Row):
            row = row_or_id
        elif row_or_id is not None:
            row = table(row_or_id, **filters)
        elif filters:
            row = table(**filters)
        else:
            # dummy object
            return inst

        if not row:
            return None  # type: ignore

        inst._row = row

        if hasattr(row, "id"):
            inst.__dict__.update(row)
        else:
            # deal with _extra (and possibly others?)
            # Row <{actual: {}, _extra: ...}>
            inst.__dict__.update(row[str(cls)])

        inst._setup_instance_methods()
        return inst

    def __iter__(self) -> t.Generator[t.Any, None, None]:
        """
        Allows looping through the columns.
        """
        row = self._ensure_matching_row()
        yield from iter(row)

    def __getitem__(self, item: str) -> t.Any:
        """
        Allows dictionary notation to get columns.
        """
        if item in self.__dict__:
            return self.__dict__.get(item)

        # fallback to lookup in row
        if self._row:
            return self._row[item]

        # nothing found!
        raise KeyError(item)

    def __getattr__(self, item: str) -> t.Any:
        """
        Allows dot notation to get columns.
        """
        if value := self.get(item):
            return value

        raise AttributeError(item)

    def __eq__(self, other: t.Any) -> bool:
        """
        Compare equal classes via their _row, since the other data is irrelevant for equality.
        """
        if type(self) is not type(other):
            return False

        return self._row == other._row  # type: ignore

    def keys(self) -> list[str]:
        """
        Return the combination of row + relationship keys.

        Used by dict(row).
        """
        return list(self._row.keys() if self._row else ()) + getattr(self, "_with", [])

    def get(self, item: str, default: t.Any = None) -> t.Any:
        """
        Try to get a column from this instance, else return default.
        """
        try:
            return self.__getitem__(item)
        except KeyError:
            return default

    def __setitem__(self, key: str, value: t.Any) -> None:
        """
        Data can both be updated via dot and dict notation.
        """
        return setattr(self, key, value)

    def __int__(self) -> int:
        """
        Calling int on a model instance will return its id.
        """
        return getattr(self, "id", 0)

    def __bool__(self) -> bool:
        """
        If the instance has an underlying row with data, it is truthy.
        """
        return bool(getattr(self, "_row", False))

    def _ensure_matching_row(self) -> Row:
        row = getattr(self, "_row", None)
        return t.cast(Row, row) or throw(
            EnvironmentError("Trying to access non-existant row. Maybe it was deleted or not yet initialized?")
        )

    def __repr__(self) -> str:
        """
        String representation of the model instance.
        """
        model_name = self.__class__.__name__
        model_data = {}

        if self._row:
            model_data = self._row.as_json()

        details = model_name
        details += f"({model_data})"

        if relationships := getattr(self, "_with", []):
            details += f" + {relationships}"

        return f"<{details}>"

    # serialization
    # underscore variants work for class instances (set up by _setup_instance_methods)

    @classmethod
    def as_dict(cls, flat: bool = False, sanitize: bool = True) -> AnyDict:
        """
        Dump the object to a plain dict.

        Can be used as both a class or instance method:
        - dumps the table info if it's a class
        - dumps the row info if it's an instance (see _as_dict)
        """
        table = cls._ensure_table_defined()
        result = table.as_dict(flat, sanitize)
        return t.cast(AnyDict, result)

    @classmethod
    def as_json(cls, sanitize: bool = True, indent: t.Optional[int] = None, **kwargs: t.Any) -> str:
        """
        Dump the object to json.

        Can be used as both a class or instance method:
        - dumps the table info if it's a class
        - dumps the row info if it's an instance (see _as_json)
        """
        data = cls.as_dict(sanitize=sanitize)
        return as_json.encode(data, indent=indent, **kwargs)

    @classmethod
    def as_xml(cls, sanitize: bool = True) -> str:  # pragma: no cover
        """
        Dump the object to xml.

        Can be used as both a class or instance method:
        - dumps the table info if it's a class
        - dumps the row info if it's an instance (see _as_xml)
        """
        table = cls._ensure_table_defined()
        return t.cast(str, table.as_xml(sanitize))

    @classmethod
    def as_yaml(cls, sanitize: bool = True) -> str:
        """
        Dump the object to yaml.

        Can be used as both a class or instance method:
        - dumps the table info if it's a class
        - dumps the row info if it's an instance (see _as_yaml)
        """
        table = cls._ensure_table_defined()
        return t.cast(str, table.as_yaml(sanitize))

    def _as_dict(
        self,
        datetime_to_str: bool = False,
        custom_types: t.Iterable[type] | type | None = None,
    ) -> AnyDict:
        row = self._ensure_matching_row()

        result = row.as_dict(datetime_to_str=datetime_to_str, custom_types=custom_types)

        def asdict_method(obj: t.Any) -> t.Any:  # pragma: no cover
            if hasattr(obj, "_as_dict"):  # typedal
                return obj._as_dict()
            elif hasattr(obj, "as_dict"):  # pydal
                return obj.as_dict()
            else:  # something else??
                return obj.__dict__

        if _with := getattr(self, "_with", None):
            for relationship in _with:
                data = self.get(relationship)

                if isinstance(data, list):
                    data = [asdict_method(_) for _ in data]
                elif data:
                    data = asdict_method(data)

                result[relationship] = data

        return t.cast(AnyDict, result)

    def _as_json(
        self,
        default: t.Callable[[t.Any], t.Any] = None,
        indent: t.Optional[int] = None,
        **kwargs: t.Any,
    ) -> str:
        data = self._as_dict()
        return as_json.encode(data, default=default, indent=indent, **kwargs)

    def _as_xml(self, sanitize: bool = True) -> str:  # pragma: no cover
        row = self._ensure_matching_row()
        return t.cast(str, row.as_xml(sanitize))

    # def _as_yaml(self, sanitize: bool = True) -> str:
    #     row = self._ensure_matching_row()
    #     return t.cast(str, row.as_yaml(sanitize))

    def __setattr__(self, key: str, value: t.Any) -> None:
        """
        When setting a property on a Typed Table model instance, also update the underlying row.
        """
        if self._row and key in self._row.__dict__ and not callable(value):
            # enables `row.key = value; row.update_record()`
            self._row[key] = value

        super().__setattr__(key, value)

    @classmethod
    def update(cls: t.Type[T_MetaInstance], query: Query, **fields: t.Any) -> T_MetaInstance | None:
        """
        Update one record.

        Example:
            MyTable.update(MyTable.id == 1, name="NewName") -> MyTable
        """
        # todo: update multiple?
        if record := cls(query):
            return record.update_record(**fields)
        else:
            return None

    def _update(self: T_MetaInstance, **fields: t.Any) -> T_MetaInstance:
        row = self._ensure_matching_row()
        row.update(**fields)
        self.__dict__.update(**fields)
        return self

    def _update_record(self: T_MetaInstance, **fields: t.Any) -> T_MetaInstance:
        row = self._ensure_matching_row()
        new_row = row.update_record(**fields)
        self.update(**new_row)
        return self

    def update_record(self: T_MetaInstance, **fields: t.Any) -> T_MetaInstance:  # pragma: no cover
        """
        Here as a placeholder for _update_record.

        Will be replaced on instance creation!
        """
        return self._update_record(**fields)

    def _delete_record(self) -> int:
        """
        Actual logic in `pydal.helpers.classes.RecordDeleter`.
        """
        row = self._ensure_matching_row()
        result = row.delete_record()
        self.__dict__ = {}  # empty self, since row is no more.
        self._row = None  # just to be sure
        self._setup_instance_methods()
        # ^ instance methods might've been deleted by emptying dict,
        # but we still want .as_dict to show an error, not the table's as_dict.
        return t.cast(int, result)

    def delete_record(self) -> int:  # pragma: no cover
        """
        Here as a placeholder for _delete_record.

        Will be replaced on instance creation!
        """
        return self._delete_record()

    # __del__ is also called on the end of a scope so don't remove records on every del!!

    # pickling:

    def __getstate__(self) -> AnyDict:
        """
        State to save when pickling.

        Prevents db connection from being pickled.
        Similar to as_dict but without changing the data of the relationships (dill does that recursively)
        """
        row = self._ensure_matching_row()
        result: AnyDict = row.as_dict()

        if _with := getattr(self, "_with", None):
            result["_with"] = _with
            for relationship in _with:
                data = self.get(relationship)

                result[relationship] = data

        result["_row"] = self._row.as_json() if self._row else ""
        return result

    def __setstate__(self, state: AnyDict) -> None:
        """
        Used by dill when loading from a bytestring.
        """
        # as_dict also includes table info, so dump as json to only get the actual row data
        # then create a new (more empty) row object:
        state["_row"] = Row(json.loads(state["_row"]))
        self.__dict__ |= state

    @classmethod
    def _sql(cls) -> str:
        """
        Generate SQL Schema for this table via pydal2sql (if 'migrations' extra is installed).
        """
        try:
            import pydal2sql
        except ImportError as e:  # pragma: no cover
            raise RuntimeError("Can not generate SQL without the 'migration' extra or `pydal2sql` installed!") from e

        return pydal2sql.generate_sql(cls)

    def render(self, fields: list[Field] = None, compact: bool = False) -> t.Self:
        """
        Renders a copy of the object with potentially modified values.

        Args:
            fields: A list of fields to render. Defaults to all representable fields in the table.
            compact: Whether to return only the value of the first field if there is only one field.

        Returns:
            A copy of the object with potentially modified values.
        """
        row = copy.deepcopy(self)
        keys = list(row)
        if not fields:
            fields = [self._table[f] for f in self._table._fields]
            fields = [f for f in fields if isinstance(f, Field) and f.represent]

        for field in fields:
            if field._table == self._table:
                row[field.name] = self._db.represent(
                    "rows_render",
                    field,
                    row[field.name],
                    row,
                )
            # else: relationship, different logic:

        for relation_name in getattr(row, "_with", []):
            if relation := self._relationships.get(relation_name):
                relation_table = relation.table
                if isinstance(relation_table, str):
                    relation_table = self._db[relation_table]

                relation_row = row[relation_name]

                if isinstance(relation_row, list):
                    # list of rows
                    combined = []

                    for related_og in relation_row:
                        related = copy.deepcopy(related_og)
                        for fieldname in related:
                            field = relation_table[fieldname]
                            related[field.name] = self._db.represent(
                                "rows_render",
                                field,
                                related[field.name],
                                related,
                            )
                        combined.append(related)

                    row[relation_name] = combined
                else:
                    # 1 row
                    for fieldname in relation_row:
                        field = relation_table[fieldname]
                        row[relation_name][fieldname] = self._db.represent(
                            "rows_render",
                            field,
                            relation_row[field.name],
                            relation_row,
                        )

        if compact and len(keys) == 1 and keys[0] != "_extra":  # pragma: no cover
            return t.cast(t.Self, row[keys[0]])
        return row


# backwards compat:
TypedRow = TypedTable

# note: at the bottom to prevent circular import issues:
from .fields import TypedField  # noqa: E402
from .query_builder import QueryBuilder  # noqa: E402
